"""Special methods (overrides) that add behavior to the model that cannot
simply be generated.

Derived methods always return a list. Note this is not the case for
normal properties.
"""

from __future__ import annotations

import itertools

from gaphor.core.modeling.properties import derived
from gaphor.diagram.group import self_and_owners
from gaphor.UML import uml, umllex


def named_element_qualified_name(self) -> list[str]:
    """Returns the qualified name of the element as a list."""

    qname = [getattr(e, "name", "??") or "??" for e in self_and_owners(self)]
    qname.reverse()
    return qname


uml.NamedElement.qualifiedName = property(named_element_qualified_name)


# See https://www.omg.org/spec/UML/2.5/PDF, section 12.4.1.5, page 271
def extension_metaclass(self):
    """References the Class that is extended through an Extension.

    The property is derived from the type of the memberEnd that is not
    the ownedEnd.
    """
    ownedEnd = self.ownedEnd
    if metaend := [e for e in self.memberEnd if e is not ownedEnd]:
        return metaend[0].type


# Don't use derived() now, it can not deal with a [0..1] property derived from a [0..*] property.
# Extension.metaclass = derived(Extension, 'metaclass', Class, 0, 1, Extension.ownedEnd, Association.memberEnd)  # noqa: E800
# Extension.metaclass.filter = extension_metaclass  # noqa: E800
uml.Extension.metaclass = property(extension_metaclass, doc=extension_metaclass.__doc__)


def property_opposite(self: uml.Property) -> list[uml.Property | None]:
    """In the case where the property is one navigable end of a binary
    association with both ends navigable, this gives the other end.

    For Gaphor the property on the other end is returned regardless the
    navigability.
    """
    if self.association is not None and len(self.association.memberEnd) == 2:
        return [
            self.association.memberEnd[1]
            if self.association.memberEnd[0] is self
            else self.association.memberEnd[0]
        ]
    return [None]


uml.Property.opposite = derived("opposite", uml.Property, 0, 1, property_opposite)


def property_navigability(self: uml.Property) -> list[bool | None]:
    """Get navigability of an association end.

    If no association is related to the property, then unknown
    navigability is assumed.
    """
    assoc = self.association
    if not (assoc and self.opposite):
        return [None]  # assume unknown
    owner = self.opposite.type
    if (
        isinstance(owner, uml.Class | uml.DataType | uml.Interface)
        and isinstance(self.type, uml.Class | uml.DataType | uml.Interface)
        and (self in owner.ownedAttribute)
        or self in assoc.navigableOwnedEnd
    ):
        return [True]
    elif isinstance(assoc.ownedEnd, uml.ExtensionEnd):
        return [self is assoc.ownedEnd]
    elif assoc.ownedEnd is None or self in assoc.ownedEnd:
        return [None]
    return [False]


uml.Property.navigability = derived("navigability", bool, 0, 1, property_navigability)


def _pr_interface_deps(classifier: uml.NamedElement, dep_type):
    """Return all interfaces, which are connected to a classifier with given
    dependency type."""
    return (
        dep.supplier
        for dep in classifier.clientDependency
        if dep.isKindOf(dep_type)
        and len(dep.supplier) > 0
        and dep.supplier[0].isKindOf(uml.Interface)
    )


def _pr_rc_interface_deps(component: uml.Component, dep_type):
    """Return all interfaces, which are connected to realizing classifiers of
    specified component. Returned interfaces are connected to realizing
    classifiers with given dependency type.

    Generator of generators is returned. Do not forget to flat it later.
    """
    return (
        _pr_interface_deps(r.realizingClassifier, dep_type)
        for r in component.realization
    )


def component_provided(self: uml.Component) -> list[uml.Realization]:
    """Interfaces provided to component environment."""
    implementations = (
        impl.contract
        for impl in self.interfaceRealization
        if impl.isKindOf(uml.InterfaceRealization)
    )
    realizations = _pr_interface_deps(self, uml.Realization)

    # realizing classifiers realizations
    # this is a generator of generators, so flatten it later
    rc_realizations = _pr_rc_interface_deps(self, uml.Realization)

    return list(set(itertools.chain(implementations, realizations, *rc_realizations)))  # type: ignore[arg-type]


uml.Component.provided = property(component_provided, doc=component_provided.__doc__)


def component_required(self) -> list[uml.Usage]:
    """Interfaces required by component."""
    usages = _pr_interface_deps(self, uml.Usage)

    # realizing classifiers usages
    # this generator of generators, so flatten it later
    rc_usages = _pr_rc_interface_deps(self, uml.Usage)

    return list(itertools.chain(usages, *rc_usages))


uml.Component.required = property(component_required, doc=component_required.__doc__)


def message_messageKind(self) -> str:
    """MessageKind."""
    kind = "unknown"
    if self.sendEvent:
        kind = "complete" if self.receiveEvent else "lost"
    elif self.receiveEvent:
        kind = "found"
    return kind


uml.Message.messageKind = property(message_messageKind, doc=message_messageKind.__doc__)


uml.Lifeline.parse = umllex.parse_lifeline
uml.Lifeline.render = umllex.render_lifeline



from pathlib import Path

from gi.repository import GLib, Gtk

from gaphor.abc import Service
from gaphor.core import event_handler
from gaphor.event import ModelSaved, SessionCreated
from gaphor.ui import APPLICATION_ID


class RecentFiles(Service):
    def __init__(self, event_manager, recent_manager=None):
        self.event_manager = event_manager
        self.recent_manager = recent_manager or Gtk.RecentManager.get_default()

        self.event_manager.subscribe(self._on_filename_changed)

    def shutdown(self):
        self.event_manager.unsubscribe(self._on_filename_changed)

    @event_handler(SessionCreated, ModelSaved)
    def _on_filename_changed(self, event):
        filename = event.filename
        if not filename:
            return
        uri = GLib.filename_to_uri(str(Path(filename).absolute()))
        # Re-add, to ensure it's at the top of the list
        self.remove(uri)
        self.add(uri)

    def add(self, uri):
        meta = Gtk.RecentData()
        meta.app_name = APPLICATION_ID
        meta.app_exec = f"{GLib.get_prgname()} %u"
        meta.mime_type = "application/x-gaphor"
        self.recent_manager.add_full(uri, meta)

    def remove(self, uri):
        # From: https://gitlab.gnome.org/GNOME/pitivi/-/blob/58b5e3b6/pitivi/application.py#L271
        try:
            self.recent_manager.remove_item(uri)
        except GLib.Error as e:
            if e.domain != "gtk-recent-manager-error-quark":
                raise e



# ruff: noqa: F401

from gaphor.C4Model.diagramitems.container import C4ContainerItem
from gaphor.C4Model.diagramitems.database import C4DatabaseItem
from gaphor.C4Model.diagramitems.dependency import C4DependencyItem
from gaphor.C4Model.diagramitems.person import C4PersonItem



from gaphor.UML import Package
from gaphor.UML.recipes import owner_package


def named_element_config(new_item, name=None):
    new_item.subject.name = new_item.diagram.gettext("New {name}").format(
        name=name or new_item.diagram.gettext(type(new_item.subject).__name__)
    )


def default_namespace(new_item):
    if not new_item.subject.namespace:
        if isinstance(new_item.subject, Package):
            new_item.subject.nestingPackage = owner_package(new_item.diagram)
        else:
            new_item.subject.package = owner_package(new_item.diagram)


def namespace_config(new_item, name=None):
    named_element_config(new_item, name)
    default_namespace(new_item)



from __future__ import annotations

from collections.abc import Iterable
from operator import setitem

from gaphor.core.modeling import (
    Base,
    ElementChange,
    ElementFactory,
    Presentation,
    RefChange,
    StyleSheet,
    UnlimitedNatural,
    ValueChange,
)
from gaphor.core.modeling.collection import collection


class UnmatchableModel(Exception):
    def __init__(self, ancestor, incoming):
        super().__init__(f"Incompatible types {ancestor} != {incoming}")
        self.ancestor = ancestor
        self.incoming = incoming


def compare(
    current: ElementFactory, ancestor: ElementFactory, incoming: ElementFactory
) -> Iterable[ElementChange | ValueChange | RefChange]:
    """Compare two models.

    Changes are recorded in the current model as `PendingChange` objects
    (`ElementChange`, `ValueChange`, `RefChange`).

    Returns an iterable of the added change objects.
    """
    ancestor_keys = set(ancestor.keys())
    incoming_keys = set(incoming.keys())

    ancestor_style_sheet = None
    incoming_style_sheet = None

    def create(type, **kwargs):
        e = current.create(type)
        for name, value in kwargs.items():
            setattr(e, name, None if value is None else str(value))
        return e

    for key in ancestor_keys.difference(incoming_keys):
        e = ancestor[key]
        if isinstance(e, StyleSheet):
            ancestor_style_sheet = e
        else:
            yield create(
                ElementChange,
                op="remove",
                element_name=type(e).__name__,
                modeling_language=type(e).__modeling_language__,
                element_id=key,
            )

    for key in incoming_keys.difference(ancestor_keys):
        e = incoming[key]
        if isinstance(e, StyleSheet):
            incoming_style_sheet = e
        else:
            yield create(
                ElementChange,
                op="add",
                element_name=type(e).__name__,
                modeling_language=type(e).__modeling_language__,
                element_id=key,
                diagram_id=e.diagram.id if isinstance(e, Presentation) else None,
            )
            yield from updated_properties(None, e, create)

    for key in ancestor_keys.intersection(incoming_keys):
        a = ancestor[key]
        i = incoming[key]
        if type(a) is not type(i):
            raise UnmatchableModel(a, i)
        yield from updated_properties(a, i, create)

    if (
        ancestor_style_sheet
        and incoming_style_sheet
        and ancestor_style_sheet.id != incoming_style_sheet.id
    ):
        yield from updated_properties(
            ancestor_style_sheet, incoming_style_sheet, create
        )


def updated_properties(ancestor, incoming, create) -> Iterable[ValueChange | RefChange]:
    ancestor_vals: dict[str, Base | collection[Base] | str | int | None] = {}
    if ancestor:
        ancestor.save(lambda n, v: setitem(ancestor_vals, n, v))
    incoming_vals: dict[str, Base | collection[Base] | str | int | None] = {}
    incoming.save(lambda n, v: setitem(incoming_vals, n, v))

    for name in {*ancestor_vals.keys(), *incoming_vals.keys()}:
        if name == "id":
            continue
        value = incoming_vals.get(name)
        other = ancestor_vals.get(name)
        id = ancestor.id if ancestor else incoming.id
        if isinstance(value, Base):
            # Allow values to be None
            assert other is None or isinstance(other, Base)
            if other is None or value.id != other.id:
                yield create(
                    RefChange,
                    op="update",
                    element_id=id,
                    property_name=name,
                    property_ref=value.id,
                )
        elif isinstance(value, collection):
            assert other is None or isinstance(other, collection)
            other_ids = {o.id for o in other} if other is not None else set()

            yield from (
                create(
                    RefChange,
                    op="add",
                    element_id=id,
                    property_name=name,
                    property_ref=v.id,
                )
                for v in value
                if v.id not in other_ids
            )
        elif value != other:
            if isinstance(other, Base):
                yield create(
                    RefChange,
                    op="update",
                    element_id=id,
                    property_name=name,
                    property_ref=None,
                )
            elif not isinstance(other, collection):
                value_change = create(
                    ValueChange,
                    op="update",
                    element_id=id,
                    property_name=name,
                )
                set_value_change_property_value(value_change, value)
                yield value_change

        if isinstance(other, collection):
            assert value is None or isinstance(value, collection)
            value_ids = {v.id for v in value} if value else set()

            yield from (
                create(
                    RefChange,
                    op="remove",
                    element_id=id,
                    property_name=name,
                    property_ref=o.id,
                )
                for o in other
                if o.id not in value_ids
            )


def set_value_change_property_value(
    value_change: ValueChange, new_value: None | str | int | UnlimitedNatural | bool
):
    if new_value is None:
        value_change.property_value = None
        value_change.property_type = None
    elif isinstance(new_value, str):
        value_change.property_value = new_value
        value_change.property_type = "str"
    elif isinstance(new_value, bool):
        value_change.property_value = str(new_value)
        value_change.property_type = "bool"
    elif isinstance(new_value, int):
        value_change.property_value = str(new_value)
        value_change.property_type = "int"
    elif new_value == "*":
        value_change.property_value = str(new_value)
        value_change.property_type = "UnlimitedNatural"



from gaphor.UML.general import connectors as connectors
from gaphor.UML.general.comment import CommentItem as CommentItem
from gaphor.UML.general.comment import CommentLineItem as CommentLineItem
from gaphor.UML.general.diagramitem import DiagramItem as DiagramItem
from gaphor.UML.general.image import ImageItem as ImageItem

__all__ = ["CommentItem", "CommentLineItem", "DiagramItem", "ImageItem"]



import importlib.resources
import logging
import platform
import sys
import textwrap
import time

import cairo
import gi
from gi.repository import Adw, Gdk, Gio, GLib, Gtk, GtkSource, Pango

from gaphor.abc import Service
from gaphor.application import Application, distribution
from gaphor.core import Transaction
from gaphor.core.modeling import Diagram
from gaphor.ui import APPLICATION_ID

log = logging.getLogger(__name__)


class Status:
    def __init__(self, name):
        self.name = name
        self.status = "in progress"

    def passed(self):
        self.status = "passed"

    def skip(self):
        self.status = "skipped"

    @property
    def in_progress(self):
        return self.status == "in progress"

    @property
    def completed(self):
        return self.status in ("passed", "skipped")

    def __repr__(self):
        return f"{self.name}: {self.status}"


def test(func):
    """A test function."""

    def wrapper(self):
        status = Status(func.__name__)
        self.statuses.append(status)
        try:
            return func(self, status)
        except BaseException:
            log.exception("Test %s failed", func.__name__)
            status.status = "failed"

    return wrapper


class SelfTest(Service):
    def __init__(self, application: Application):
        self.application = application
        self.statuses: list[Status] = []

    def shutdown(self):
        pass

    def init(self, gtk_app):
        windows_console_output_workaround()
        self.init_timer(gtk_app, timeout=30)
        self.test_library_versions()
        self.test_gsettings_schemas()
        self.test_new_session()
        self.test_auto_layout()

    def init_timer(self, gtk_app, timeout):
        start = time.time()

        def callback():
            if time.time() > start + timeout:
                log.error("Tests timed out")
                gtk_app.exit_code = 1
            elif any(status.in_progress for status in self.statuses):
                return GLib.SOURCE_CONTINUE
            elif all(status.completed for status in self.statuses):
                log.info(
                    "All tests have been completed in %.1fs",
                    time.time() - start,
                )
            else:
                log.error("Not all tests have passed")
                gtk_app.exit_code = 1

            for status in self.statuses:
                log.log(logging.INFO if status.completed else logging.ERROR, status)

            gtk_app.quit()
            return GLib.SOURCE_REMOVE

        GLib.timeout_add(priority=GLib.PRIORITY_LOW, interval=100, function=callback)

    @test
    def test_library_versions(self, status):
        log.info(
            "System information:\n\n%s", textwrap.indent(system_information(), "\t")
        )
        status.passed()

    @test
    def test_new_session(self, status):
        def check_new_session(session):
            main_window = session.get_service("main_window")

            if main_window.window and main_window.window.get_visible():
                status.passed()
                return GLib.SOURCE_REMOVE
            return GLib.SOURCE_CONTINUE

        template = importlib.resources.files("gaphor") / "templates" / "uml.gaphor"
        session = self.application.new_session(template=template)
        GLib.idle_add(check_new_session, session, priority=GLib.PRIORITY_LOW)

    @test
    def test_gsettings_schemas(self, status):
        source = Gio.settings_schema_source_get_default()
        data_dirs = [GLib.get_user_data_dir(), *GLib.get_system_data_dirs()]
        for schema in ["org.gtk.gtk4.Settings.FileChooser", APPLICATION_ID]:
            if source.lookup(schema, recursive=True):
                log.info("Schema %s found", schema)
            else:
                log.debug("Schemas found: %s %s", *source.list_schemas(True))
                raise RuntimeError(
                    f"Could not find schema {schema} in data dirs: {':'.join(data_dirs)}"
                )
        status.passed()

    @test
    def test_auto_layout(self, status):
        session = self.application.new_session()
        event_manager = session.get_service("event_manager")
        element_factory = session.get_service("element_factory")
        auto_layout = session.get_service("auto_layout")

        with Transaction(event_manager):
            diagram = element_factory.create(Diagram)

        auto_layout.layout(diagram)
        status.passed()


def system_information():
    return textwrap.dedent(
        f"""\
        Gaphor version:         {distribution().version}
        Operating System:       {platform.system()} ({platform.release()}, {platform.machine()})
        Display:                {display_type()}
        Python version:         {platform.python_version()}
        GTK version:            {Gtk.get_major_version()}.{Gtk.get_minor_version()}.{Gtk.get_micro_version()}
        Adwaita version:        {Adw.get_major_version()}.{Adw.get_minor_version()}.{Adw.get_micro_version()}
        GtkSourceView version:  {gtk_source_view_version()}
        Cairo version:          {cairo.cairo_version_string()}
        Pango version:          {Pango.version_string()}
        PyGObject version:      {gi.__version__}
        Pycairo version:        {cairo.version}
        """
    )


def display_type():
    dm = Gdk.DisplayManager.get()
    display = dm.get_default_display()
    return display.__class__.__name__ if display else "none"


def gtk_source_view_version():
    if hasattr(GtkSource, "get_major_version"):
        return f"{GtkSource.get_major_version()}.{GtkSource.get_minor_version()}.{GtkSource.get_micro_version()}"
    return "-"


def windows_console_output_workaround():
    if sys.platform == "win32":
        from gaphor.main import LOG_FORMAT

        logging.basicConfig(
            level=logging.INFO,
            format=LOG_FORMAT,
            filename="gaphor-self-test.txt",
            filemode="w",
            force=True,
            encoding="utf-8",
        )



"""The Sanitize module is dedicated to adapters (stuff) that keeps the model
clean and in sync with diagrams."""

from collections.abc import Iterable

from gaphor import UML
from gaphor.abc import Service
from gaphor.core import event_handler
from gaphor.core.modeling import Diagram, Presentation
from gaphor.core.modeling.event import (
    AssociationDeleted,
    AssociationSet,
    DerivedSet,
    DiagramUpdateRequested,
)
from gaphor.diagram.deletable import deletable
from gaphor.event import Notification
from gaphor.i18n import gettext
from gaphor.settings import settings
from gaphor.transaction import TransactionBegin, TransactionCommit, TransactionRollback
from gaphor.UML.general.comment import CommentLineItem


def undo_guard(func):
    """Do not execute the sanitizer if we're undoing/redoing/rolling back a transaction.

    The sanitizer actions are already part of that transaction.
    """

    def guard(self, event):
        if self.in_undo_transaction:
            return
        return func(self, event)

    return guard


class SanitizerService(Service):
    """Does some background cleanup jobs, such as removing elements from the
    model that have no presentations (and should have some)."""

    def __init__(self, event_manager, properties=None):
        self.event_manager = event_manager
        self.properties = properties or {}
        self.in_undo_transaction = False
        self._to_be_updated_diagrams = set()

        event_manager.subscribe(self._begin_transaction)
        event_manager.subscribe(self._end_transaction)
        event_manager.subscribe(self._diagram_update_requested)
        event_manager.subscribe(self._on_update_diagrams)
        event_manager.subscribe(self._unlink_on_subject_delete)
        event_manager.subscribe(self._update_annotated_element_link)
        event_manager.subscribe(self._unlink_on_extension_delete)
        event_manager.subscribe(self._redraw_diagram_on_move)

    def shutdown(self):
        event_manager = self.event_manager
        event_manager.unsubscribe(self._begin_transaction)
        event_manager.unsubscribe(self._end_transaction)
        event_manager.unsubscribe(self._on_update_diagrams)
        event_manager.unsubscribe(self._diagram_update_requested)
        event_manager.unsubscribe(self._unlink_on_subject_delete)
        event_manager.unsubscribe(self._update_annotated_element_link)
        event_manager.unsubscribe(self._unlink_on_extension_delete)
        event_manager.unsubscribe(self._redraw_diagram_on_move)

    @event_handler(TransactionBegin)
    def _begin_transaction(self, event):
        self.in_undo_transaction = bool(event.context)

    @event_handler(TransactionCommit, TransactionRollback)
    def _end_transaction(self, _event):
        self.in_undo_transaction = False

    @event_handler(DiagramUpdateRequested)
    def _diagram_update_requested(self, event: DiagramUpdateRequested):
        self._to_be_updated_diagrams.add(event.diagram)

    @event_handler(TransactionCommit, TransactionRollback)
    def _on_update_diagrams(self, event):
        for diagram in self._to_be_updated_diagrams:
            diagram.update()
        self._to_be_updated_diagrams.clear()

    @event_handler(AssociationSet)
    @undo_guard
    def _unlink_on_subject_delete(self, event):
        """Unlink the model element if no more presentations link to the
        `item`'s subject or the deleted item was the only item currently
        linked."""
        if (
            not settings.remove_unused_elements
            or event.property is not Presentation.subject  # type: ignore[misc]
        ):
            return

        if old_subject := event.old_value:
            if isinstance(old_subject, UML.Package) and old_subject.ownedElement:
                return
            elif deletable(old_subject) and not old_subject.presentation:
                old_subject.unlink()

                self.event_manager.handle(
                    Notification(
                        gettext(
                            "Removed unused elements from the model: they are not used in any other diagram."
                        )
                    )
                )

    @event_handler(AssociationSet)
    @undo_guard
    def _update_annotated_element_link(self, event):
        """Link comment and element if a comment line is present, but comment
        and element subject are not connected yet."""
        if event.property is not Presentation.subject:  # type: ignore[misc]
            return

        element: Presentation = event.element
        subject = event.new_value
        if not (element.diagram and subject):
            return

        for cinfo in element.diagram.connections.get_connections(connected=element):
            comment_line = cinfo.item
            if not isinstance(comment_line, CommentLineItem):
                continue
            opposite = comment_line.opposite(cinfo.handle)
            opposite_cinfo = element.diagram.connections.get_connection(opposite)
            if not opposite_cinfo:
                continue
            comment_item = opposite_cinfo.connected
            comment = comment_item.subject
            comment.annotatedElement = subject

    @event_handler(AssociationDeleted)
    @undo_guard
    def _unlink_on_extension_delete(self, event):
        """Remove applied stereotypes when extension is deleted."""
        if isinstance(event.element, UML.Stereotype) and event.property in (
            UML.Stereotype.ownedAttribute,
            UML.Stereotype.generalization,
        ):
            update_stereotype_application(event.element)

    @event_handler(DerivedSet)
    @undo_guard
    def _redraw_diagram_on_move(self, event):
        if event.property is UML.Element.owner and isinstance(event.element, Diagram):
            diagram = event.element
            for item in diagram.get_all_items():
                diagram.request_update(item)


def update_stereotype_application(stereotype: UML.Stereotype, seen=None):
    # Seen is a (mutable) list to avoid infinite loops
    if seen is None:
        seen = {stereotype}

    metaclass_names = {m.name for m in metaclasses(stereotype)}

    for instance_spec in list(stereotype.instanceSpecification):
        for e in list(instance_spec.extended):
            names = {c.__name__ for c in type(e).__mro__ if issubclass(c, UML.Element)}
            if names.isdisjoint(metaclass_names):
                del instance_spec.extended[e]
        if not instance_spec.extended:
            instance_spec.unlink()

    for sub_stereotype in stereotype.specialization[:].specific:
        if sub_stereotype not in seen:
            seen.add(sub_stereotype)
            update_stereotype_application(sub_stereotype, seen)


def metaclasses(stereotype: UML.Stereotype, seen=None) -> Iterable[UML.Class]:
    # Seen is a (mutable) list to avoid infinite loops
    if seen is None:
        seen = {stereotype}

    extensions = (
        e
        for e in stereotype.ownedAttribute[:].association
        if isinstance(e, UML.Extension)
    )

    yield from (e.metaclass for e in extensions if e.metaclass)

    for super_stereotype in stereotype.generalization[:].general:
        if super_stereotype not in seen:
            seen.add(super_stereotype)
            yield from metaclasses(super_stereotype, seen)



"""Internationalization (i18n) support for Gaphor.

Translate text in to your native language using the gettext() function.
"""

from __future__ import annotations

import functools
import gettext as _gettext
import importlib.resources
import locale
import logging
import os
import sys

import defusedxml.ElementTree as etree

log = logging.getLogger(__name__)

localedir = importlib.resources.files("gaphor") / "locale"


def _get_os_language() -> str:
    """Get the default language in Windows or macOS.

    Inspired by https://github.com/nicotine-plus/nicotine-plus
    locale.getlocale() fails to get the correct language if the region
    is set different from the language.
    """
    if sys.platform == "darwin":
        from Cocoa import NSUserDefaults

        defaults = NSUserDefaults.standardUserDefaults()
        langs = defaults.objectForKey_("AppleLanguages")
        if language := langs.objectAtIndex_(0):
            assert isinstance(language, str)
            return language.replace("-", "_")
    elif sys.platform == "win32":
        import ctypes

        windll = ctypes.windll.kernel32
        if language := locale.windows_locale.get(windll.GetUserDefaultUILanguage()):
            return language
    return ""


@functools.lru_cache(maxsize=2)
def translation(lang) -> _gettext.GNUTranslations | _gettext.NullTranslations:
    try:
        return _gettext.translation(
            "gaphor", localedir=str(localedir), languages=[lang, "C"]
        )
    except FileNotFoundError as e:
        if lang.lower() not in ("c", "c.utf-8", "en_us", "en_us.utf-8"):
            log.warning(f"No translations were found for language {lang}: {e}")
    return _gettext.NullTranslations()


def force_english_locale():
    """Force English locale, instead of OS language."""
    global gettext
    gettext = translation("en_US.UTF-8").gettext


gettext = translation(os.getenv("LANG") or _get_os_language()).gettext


def i18nize(message):
    """Pick up text for internationalization without actually translating it, like gettext() does."""
    return message


@functools.cache
def translated_ui_string(package: str, ui_filename: str) -> str:
    with (importlib.resources.files(package) / ui_filename).open(
        encoding="utf-8"
    ) as ui_file:
        ui_xml = etree.parse(ui_file)
    for node in ui_xml.findall(".//*[@translatable='yes']"):
        node.text = gettext(node.text) if node.text else ""
        del node.attrib["translatable"]
    return etree.tostring(ui_xml.getroot(), encoding="unicode", method="xml")  # type: ignore[no-any-return]



import hashlib
from pathlib import Path

from gi.repository import Adw, Gio, GLib, Gtk

from gaphor.abc import ActionProvider
from gaphor.core import event_handler
from gaphor.core.modeling import ModelReady
from gaphor.event import ModelSaved
from gaphor.i18n import translated_ui_string
from gaphor.ui.abc import UIComponent


def new_builder():
    builder = Gtk.Builder()
    builder.add_from_string(translated_ui_string("gaphor.ui", "modelchanged.ui"))
    return builder


def file_digest(filename: Path | None) -> bytes | None:
    if filename is None or not filename.exists():
        return None

    with open(filename, "rb") as f:
        return hashlib.file_digest(f, "sha256").digest()


class ModelChanged(UIComponent, ActionProvider):
    """Show a banner if the current model file has changed.

    This service checks for content changes (file digest)
    to determine if a file really changed.
    """

    def __init__(self, event_manager):
        self.event_manager = event_manager
        self._filename: Path | None = None
        self._file_digest: bytes | None = None
        self._banner: Adw.Banner | None = None
        self._monitor: Gio.FileMonitor | None = None
        self._timeout_id: int = 0
        self.event_manager.subscribe(self._on_model_reset)

    def shutdown(self):
        super().shutdown()
        self.event_manager.unsubscribe(self._on_model_reset)

    def open(self):
        builder = new_builder()
        self._banner = builder.get_object("model-changed")
        self._update_monitor()
        return self._banner

    def close(self):
        self._cancel_monitor()
        self._banner = None

    def _update_monitor(self):
        self._cancel_monitor()
        if self._filename:
            monitor = Gio.File.parse_name(str(self._filename)).monitor(
                Gio.FileMonitorFlags.NONE, None
            )
            monitor.connect("changed", self._on_file_changed)
            self._monitor = monitor

    def _cancel_monitor(self):
        if self._monitor:
            self._monitor.disconnect_by_func(self._on_file_changed)
            self._monitor = None

    def _on_file_changed(self, _monitor, _file, _other_file, event_type):
        if (
            self._timeout_id == 0
            and self._banner
            and event_type != Gio.FileMonitorEvent.ATTRIBUTE_CHANGED
            and self._file_digest != file_digest(self._filename)
        ):
            self._timeout_id = GLib.timeout_add(1000, self._delayed_reveal)

    def _delayed_reveal(self):
        if self._banner:
            self._banner.set_revealed(True)
        self._timeout_id = 0
        return GLib.SOURCE_REMOVE

    @event_handler(ModelReady, ModelSaved)
    def _on_model_reset(self, event: ModelReady | ModelSaved):
        if event.filename != self._filename:
            self._filename = event.filename
            self._update_monitor()

        self._file_digest = file_digest(self._filename)

        if self._timeout_id:
            GLib.source_remove(self._timeout_id)
            self._timeout_id = 0

        if self._banner:
            self._banner.set_revealed(False)



"""1:n and n:m relations in the data model are saved using a collection."""

from __future__ import annotations

import contextlib
from collections.abc import Iterator, Sequence
from typing import Generic, TypeVar, overload

from gaphor.core.modeling.event import AssociationUpdated

T = TypeVar("T")


class collection(Generic[T]):
    """Collection (set-like) for model elements' 1:n and n:m relationships."""

    def __init__(self, property, object, type: type[T]):
        self.property = property
        self.object = object
        self.type = type
        self.items: list[T] = []

    def __len__(self) -> int:
        return len(self.items)

    def __setitem__(self, key, value) -> None:
        raise RuntimeError("items should not be overwritten.")

    def __delitem__(self, key: T) -> None:
        self.remove(key)

    @overload
    def __getitem__(self, key: int) -> T: ...

    @overload
    def __getitem__(self, key: slice[None, None, None]) -> recurseproxy[T]: ...

    def __getitem__(self, key):
        if key == _recurseproxy_trigger:
            return recurseproxy(self.items)
        return self.items.__getitem__(key)

    def __contains__(self, obj) -> bool:
        return self.items.__contains__(obj)

    def __iter__(self) -> Iterator[T]:
        return iter(self.items)

    def __str__(self):
        return f"collection({self.items})"

    __repr__ = __str__

    def __bool__(self):
        return self.items != []

    def __eq__(self, other):
        return self.items == other or (
            isinstance(other, collection) and self.items == other.items
        )

    def __hash__(self):
        return hash(str(self))

    def index(self, key: T) -> int:
        """Given an object, return the position of that object in the
        collection."""
        return self.items.index(key)

    def append(self, value: T) -> None:
        if isinstance(value, self.type):
            self.property.set(self.object, value)
        else:
            raise TypeError(f"Object is not of type {self.type.__name__}")

    def remove(self, value: T) -> None:
        if value in self.items:
            self.property.delete(self.object, value)

    # OCL members (from SMW by Ivan Porres, http://www.abo.fi/~iporres/smw)

    def size(self):
        return len(self.items)

    def includes(self, o):
        return o in self.items

    def excludes(self, o):
        return not self.includes(o)

    def count(self, o):
        return self.items.count(o)

    def includesAll(self, c):
        return all(o in self.items for o in c)

    def excludesAll(self, c):
        return all(o not in self.items for o in c)

    def select(self, f):
        return [v for v in self.items if f(v)]

    def reject(self, f):
        return [v for v in self.items if not f(v)]

    def collect(self, f):
        return [f(v) for v in self.items]

    def isEmpty(self):
        return len(self.items) == 0

    def nonEmpty(self):
        return not self.isEmpty()

    def swap(self, item1, item2):
        """Swap two elements.

        Return true if swap was successful.
        """
        try:
            i1 = self.items.index(item1)
            i2 = self.items.index(item2)
            self.items[i1], self.items[i2] = self.items[i2], self.items[i1]
        except (IndexError, ValueError):
            return False
        else:
            self.object.handle(AssociationUpdated(self.object, self.property))
            return True

    def order(self, key):
        self.items.sort(key=key)
        self.object.handle(AssociationUpdated(self.object, self.property))


_recurseproxy_trigger = slice(None, None, None)


class recurseproxy(Generic[T]):
    """Proxy object (helper) for the recusemixin.

    The proxy has limited capabilities compared to a real list (or set):
    it can be iterated and a getitem can be performed. On the other
    side, the type of the original sequence is maintained, so getitem
    operations act as if they're executed on the original list.
    """

    def __init__(self, sequence: Sequence[T]):
        self.__sequence = sequence

    def __getitem__(self, key: int | slice[None, None, None]) -> T | Sequence[T]:
        return self.__sequence.__getitem__(key)

    def __iter__(self):
        """Iterate over the items.

        If there is some level of nesting, the parent items are iterated
        as well.
        """
        return iter(self.__sequence)

    def __getattr__(self, key: str) -> recurseproxy[T]:
        """Create a new proxy for the attribute."""

        def mygetattr():
            sentinel = object()
            for e in self.__sequence:
                obj = getattr(e, key, sentinel)
                if obj is sentinel:
                    pass
                elif issafeiterable(obj):
                    yield from obj  # type: ignore[misc]
                else:
                    yield obj

        # Create a copy of the proxy type, including a copy of the sequence type
        return type(self)(type(self.__sequence)(mygetattr()))  # type: ignore[call-arg]


def issafeiterable(obj):
    """Checks if the object is iterable, but not a string.

    >>> issafeiterable([])
    True
    >>> issafeiterable(set())
    True
    >>> issafeiterable({})
    True
    >>> issafeiterable(1)
    False
    >>> issafeiterable("text")
    False
    """
    with contextlib.suppress(TypeError):
        return iter(obj) and not isinstance(obj, str)
    return False



"""Dormant Event item definition."""

from gaphas.geometry import Rectangle

from gaphor.core.modeling import DrawContext
from gaphor.diagram.presentation import (
    Classified,
    ElementPresentation,
    text_name,
)
from gaphor.diagram.shapes import Box, IconBox, draw_diamond
from gaphor.diagram.support import represents
from gaphor.RAAML import raaml
from gaphor.RAAML.fta.undevelopedevent import draw_undeveloped_event
from gaphor.UML.compartments import text_from_package, text_stereotypes


@represents(raaml.DormantEvent)
class DormantEventItem(Classified, ElementPresentation):
    def __init__(self, diagram, id=None):
        super().__init__(diagram, id, width=70, height=35)

        self.watch("subject[UML:NamedElement].name").watch(
            "subject[UML:NamedElement].namespace.name"
        )

    def update_shapes(self, event=None):
        self.shape = IconBox(
            Box(
                draw=draw_dormant_event,
            ),
            text_stereotypes(self, lambda: [self.diagram.gettext("Dormant Event")]),
            text_name(self),
            text_from_package(self),
        )


def draw_dormant_event(box, context: DrawContext, bounding_box: Rectangle):
    draw_undeveloped_event(box, context, bounding_box)
    x1 = bounding_box.width / 5.0
    x2 = bounding_box.width * 4.0 / 5.0
    y1 = bounding_box.height / 5.0
    y2 = bounding_box.height * 4.0 / 5.0
    draw_diamond(context, x1, x2, y1, y2)



"""CommentItem diagram item."""

from gaphor.diagram.connectors import Connector
from gaphor.diagram.presentation import ElementPresentation, LinePresentation
from gaphor.diagram.shapes import Box, CssNode, Text, stroke
from gaphor.diagram.support import represents
from gaphor.UML import Comment
from gaphor.UML.compartments import text_stereotypes


@represents(Comment)
class CommentItem(ElementPresentation):
    def __init__(self, diagram, id=None):
        super().__init__(diagram, id)
        self.height = 50
        self.width = 100

        self.shape = Box(
            text_stereotypes(self),
            CssNode(
                "body",
                None,
                Text(
                    text=lambda: self.subject.body or "",
                ),
            ),
            draw=draw_border,
        )
        self.watch("subject[Comment].body")
        self.watch("subject[Element].appliedStereotype.classifier.name")


def draw_border(box, context, bounding_box):
    cr = context.cairo
    x, y, w, h = bounding_box
    _, padding_right, _, padding_left = context.style.get("padding", (4, 16, 4, 4))
    ear = max(padding_right - padding_left, padding_left)
    line_to = cr.line_to
    cr.move_to(w - ear, y)
    line_to(w - ear, y + ear)
    line_to(w, y + ear)
    line_to(w - ear, y)
    line_to(x, y)
    line_to(x, h)
    line_to(w, h)
    line_to(w, y + ear)
    stroke(context, fill=True)


class CommentLineItem(LinePresentation):
    def __init__(self, diagram, id=None):
        super().__init__(diagram, id)

    def unlink(self):
        c1 = self._connections.get_connection(self.head)
        c2 = self._connections.get_connection(self.tail)
        if c1 and c2:
            adapter = Connector(c1.connected, self)
            adapter.disconnect(self.head)
        super().unlink()



from gaphor.core.modeling import Diagram, Presentation
from gaphor.diagram.drop import drop
from gaphor.diagram.presentation import connect
from gaphor.diagram.support import get_diagram_item
from gaphor.SysML.diagramframe import DiagramFrameItem
from gaphor.SysML.sysml import Block, ProxyPort
from gaphor.UML.uml import Element, Property


@drop.register(Element, DiagramFrameItem)
def drop_element_on_diagram_frame(
    element: Element, diagram_frame: DiagramFrameItem, x, y
):
    if item := drop(element, diagram_frame.diagram, x, y):
        item.parent = diagram_frame
        item.request_update()

    return item


@drop.register(Presentation, DiagramFrameItem)
def drop_presentation_on_diagram_frame(
    item: Presentation, diagram_frame: DiagramFrameItem, x, y
):
    """When dropped on a diagram frame, change item ownership.

    Do not change model element ownership.
    """
    item.change_parent(diagram_frame)
    item.request_update()

    return item


@drop.register(ProxyPort, Diagram)
def drop_proxy_port(element: ProxyPort, diagram: Diagram, x, y):
    item_class = get_diagram_item(type(element))
    if not item_class:
        return None

    def port_owning_item_block(item):
        match item.subject:
            case Property():
                block = (
                    item.subject.type if isinstance(item.subject.type, Block) else None
                )
            case Block():
                block = item.subject
            case _:
                return False
        return element in block.ownedPort if block else False

    def drop_distance_to_item(item):
        local_x, local_y = item.matrix_i2c.inverse().transform_point(x, y)
        return item.point(local_x, local_y)

    head_item = min(
        diagram.select(lambda item: port_owning_item_block(item)),
        key=drop_distance_to_item,
        default=None,
    )

    if not head_item:
        return None

    item = diagram.create(item_class)
    assert item

    item.matrix.translate(x, y)
    item.subject = element

    connect(item, item.handles()[0], head_item)

    return item



# ruff: noqa: F401,F403

from gaphor.core.modeling.base import Base, Id, swap_element_type
from gaphor.core.modeling.coremodel import (
    ElementChange,
    PendingChange,
    RefChange,
    ValueChange,
)
from gaphor.core.modeling.diagram import (
    Diagram,
    DrawContext,
    UpdateContext,
    css_name,
)
from gaphor.core.modeling.elementfactory import ElementFactory
from gaphor.core.modeling.event import *
from gaphor.core.modeling.presentation import Presentation
from gaphor.core.modeling.properties import UnlimitedNatural
from gaphor.core.modeling.stylesheet import StyleSheet

__modeling_language__ = "Core"



from gaphor.SysML.requirements.relationships import (
    DeriveReqtItem,
    RefineItem,
    SatisfyItem,
    TraceItem,
    VerifyItem,
)
from gaphor.SysML.requirements.requirement import RequirementItem

__all__ = [
    "DeriveReqtItem",
    "RefineItem",
    "RequirementItem",
    "SatisfyItem",
    "TraceItem",
    "VerifyItem",
]



"""SEQ gate item definition."""

from gaphas.geometry import Rectangle

from gaphor.core.modeling import DrawContext
from gaphor.diagram.presentation import (
    Classified,
    ElementPresentation,
    text_name,
)
from gaphor.diagram.shapes import Box, IconBox, stroke
from gaphor.diagram.support import represents
from gaphor.RAAML import raaml
from gaphor.RAAML.fta.andgate import draw_and_gate
from gaphor.RAAML.fta.constants import DEFAULT_FTA_MAJOR, DEFAULT_FTA_MINOR
from gaphor.UML.compartments import text_from_package, text_stereotypes


@represents(raaml.SEQ)
class SEQItem(Classified, ElementPresentation):
    def __init__(self, diagram, id=None):
        super().__init__(diagram, id, width=DEFAULT_FTA_MINOR, height=DEFAULT_FTA_MAJOR)

        self.watch("subject[UML:NamedElement].name").watch(
            "subject[UML:NamedElement].namespace.name"
        )

    def update_shapes(self, event=None):
        self.shape = IconBox(
            Box(
                draw=draw_seq_gate,
            ),
            text_stereotypes(
                self, lambda: [self.diagram.gettext("Sequence Enforcing Gate")]
            ),
            text_name(self),
            text_from_package(self),
        )


def draw_seq_gate(box, context: DrawContext, bounding_box: Rectangle):
    draw_and_gate(box, context, bounding_box)
    cr = context.cairo
    wall_top = bounding_box.height / 2.0
    wall_bottom = bounding_box.height * 4.0 / 5.0

    # Triangle
    cr.move_to(0, wall_bottom)
    ry = bounding_box.height * 2.0 / 3.0
    cr.line_to(bounding_box.width / 2.0, wall_top - ry / 2.0)
    cr.line_to(bounding_box.width, wall_bottom)
    stroke(context, fill=True)



"""Inhibit gate item definition."""

from gaphas.geometry import Rectangle

from gaphor.core.modeling import DrawContext
from gaphor.diagram.presentation import (
    Classified,
    ElementPresentation,
    text_name,
)
from gaphor.diagram.shapes import Box, IconBox, stroke
from gaphor.diagram.support import represents
from gaphor.RAAML import raaml
from gaphor.RAAML.fta.constants import DEFAULT_FTA_MAJOR
from gaphor.UML.compartments import text_from_package, text_stereotypes


@represents(raaml.INHIBIT)
class InhibitItem(Classified, ElementPresentation):
    def __init__(self, diagram, id=None):
        super().__init__(diagram, id, width=DEFAULT_FTA_MAJOR, height=DEFAULT_FTA_MAJOR)

        self.watch("subject[UML:NamedElement].name").watch(
            "subject[UML:NamedElement].namespace.name"
        )

    def update_shapes(self, event=None):
        self.shape = IconBox(
            Box(
                draw=draw_inhibit_gate,
            ),
            text_stereotypes(self, lambda: [self.diagram.gettext("Inhibit Gate")]),
            text_name(self),
            text_from_package(self),
        )


def draw_inhibit_gate(box, context: DrawContext, bounding_box: Rectangle):
    cr = context.cairo

    # Top vertical line
    left = 0
    middle_width = bounding_box.width * 5.0 / 12.0
    bottom_point = bounding_box.height * 5.0 / 6.0
    top_point = bounding_box.height / 6.0
    right = bounding_box.width * 5.0 / 6.0
    upper = bounding_box.height / 3.0
    lower = bounding_box.height * 2.0 / 3.0
    cr.move_to(middle_width, 0)
    cr.line_to(middle_width, top_point)

    # Move around the hexagon counter-clockwise
    cr.line_to(left, upper)  # 1st side
    cr.line_to(left, lower)  # 2nd side
    cr.line_to(middle_width, bottom_point)  # 3rd side
    cr.line_to(right, lower)  # 4th side
    cr.line_to(right, upper)  # 5th side
    cr.line_to(middle_width, top_point)  # 6th side

    # Bottom vertical line
    cr.move_to(middle_width, bounding_box.height)
    cr.line_to(middle_width, bottom_point)

    # Right horizontal line
    middle_height = bounding_box.height / 2.0
    cr.move_to(right, middle_height)
    cr.line_to(bounding_box.width, middle_height)
    stroke(context, fill=True)



"""Grouping functionality allows nesting of one item within another item
(parent item). This is useful in several use cases.

These concepts are mostly used in the ownership module (`change_owner`)
"""

from __future__ import annotations

import itertools
from collections.abc import Callable, Iterator
from enum import Enum
from functools import singledispatch

from generic.multidispatch import FunctionDispatcher, multidispatch

from gaphor.core.modeling import Base, Diagram, Presentation


def self_and_owners(element: Base | RootType | None) -> Iterator[Base]:
    """Return the element and the ancestors (Element.owner)."""
    seen = set()
    e = element
    while isinstance(e, Base):
        if e in seen:
            return
        yield e
        seen.add(e)
        e = owner(e)


def change_owner(new_parent: Base | None, element: Base) -> bool:
    if new_parent and element.model is not new_parent.model:
        return False

    o = owner(element)
    if o is Root:
        o = None

    if new_parent and o is new_parent:
        # idempotency
        return True

    if new_parent and element in self_and_owners(new_parent):
        return False

    if new_parent is None and isinstance(o, Base):
        return ungroup(o, element)  # type: ignore[no-any-return]

    if not can_group(new_parent, element):
        return False

    if isinstance(o, Base):
        ungroup(o, element)

    return group(new_parent, element)  # type: ignore[no-any-return]


class RootType(Enum):
    Root = 1


Root = RootType.Root


@singledispatch
def owner(_element: Base | RootType | None) -> Base | RootType | None:
    return None


@owner.register
def _(element: Presentation):
    return element.diagram


@singledispatch
def owns(_element: Base) -> list[Base]:
    return []


@owns.register
def _(element: Diagram):
    return element.ownedPresentation


def no_group(parent, element) -> bool:
    return False


group: FunctionDispatcher[Callable[[Base, Base], bool]] = multidispatch(object, object)(
    no_group
)

group.register(None, object)(no_group)


def can_group(parent: Base | None, element_or_type: Base | type[Base]) -> bool:
    element_type = (
        type(element_or_type) if isinstance(element_or_type, Base) else element_or_type
    )
    parent_mro = type(parent).__mro__ if parent else [None]
    get_registration = group.registry.get_registration
    for t1, t2 in itertools.product(parent_mro, element_type.__mro__):
        if r := get_registration(t1, t2):
            return r is not no_group
    return False


@multidispatch(object, object)
def ungroup(parent, element) -> bool:
    return False


@ungroup.register(None, Base)
def none_ungroup(none, element):
    """In the rare (error?) case a model element has no parent, but is grouped
    in a diagram, allow it to ungroup."""
    return True



import math

from gi.repository import Gio

from gaphor import UML
from gaphor.diagram.propertypages import (
    LabelValue,
    PropertyPageBase,
    PropertyPages,
    handler_blocking,
    new_resource_builder,
    unsubscribe_all_on_destroy,
)
from gaphor.transaction import Transaction
from gaphor.UML.actions.activitynodes import DecisionNodeItem, ForkNodeItem
from gaphor.UML.actions.objectnode import ObjectNodeItem

new_builder = new_resource_builder("gaphor.UML.actions")


@PropertyPages.register(UML.ObjectNode)
class ObjectNodePropertyPage(PropertyPageBase):
    order = 15

    ORDERING_VALUES = ["unordered", "ordered", "LIFO", "FIFO"]

    def __init__(self, subject: UML.ObjectNode, event_manager):
        self.subject = subject
        self.event_manager = event_manager

    def construct(self):
        subject = self.subject

        if not subject or isinstance(subject, UML.ActivityParameterNode):
            return

        builder = new_builder(
            "object-node-editor",
            signals={
                "upper-bound-changed": (self._on_upper_bound_change,),
                "ordering-changed": (self._on_ordering_change,),
            },
        )

        upper_bound = builder.get_object("upper-bound")
        upperBoundText = ""
        if subject.upperBound:
            raw_upper_bound_text = UML.recipes.get_literal_value_as_string(
                subject.upperBound
            )
            if isinstance(raw_upper_bound_text, str):
                upperBoundText = raw_upper_bound_text
        upper_bound.set_text(upperBoundText)

        ordering = builder.get_object("ordering")
        ordering.set_selected(self.ORDERING_VALUES.index(subject.ordering))

        return builder.get_object("object-node-editor")

    def _on_upper_bound_change(self, entry):
        value = entry.get_text().strip()
        with Transaction(self.event_manager, context="editing"):
            if value == "":
                self.subject.upperBound = None
            else:
                upperBound = self.subject.model.create(UML.LiteralString)
                upperBound.value = value
                upperBound.objectNode = self.subject
                self.subject.upperBound = upperBound

    def _on_ordering_change(self, dropdown, _pspec):
        value = self.ORDERING_VALUES[dropdown.get_selected()]
        with Transaction(self.event_manager):
            self.subject.ordering = value


@PropertyPages.register(ObjectNodeItem)
class ShowObjectNodePropertyPage(PropertyPageBase):
    order = 16

    def __init__(self, item: ObjectNodeItem, event_manager):
        self.item = item
        self.event_manager = event_manager

    def construct(self):
        subject = self.item.subject

        if not subject:
            return

        builder = new_builder(
            "show-object-node-editor",
            signals={
                "show-ordering-changed": (self._on_ordering_show_change,),
            },
        )

        show_ordering = builder.get_object("show-ordering")
        show_ordering.set_active(self.item.show_ordering)

        return builder.get_object("show-object-node-editor")

    def _on_ordering_show_change(self, button, gparam):
        with Transaction(self.event_manager):
            self.item.show_ordering = button.get_active()


@PropertyPages.register(UML.ValueSpecificationAction)
class ValueSpecificationActionPropertyPage(PropertyPageBase):
    order = 15

    def __init__(self, item, event_manager):
        self.subject = item
        self.event_manager = event_manager

    def construct(self):
        builder = new_builder("value-specifiation-action-editor")

        value = builder.get_object("value")
        value.set_text(self.subject.value or "")
        value.connect("changed", self._on_value_change)

        return builder.get_object("value-specifiation-action-editor")

    def _on_value_change(self, entry):
        with Transaction(self.event_manager, context="editing"):
            value = entry.get_text()
            self.subject.value = value


@PropertyPages.register(UML.CallBehaviorAction)
class CallBehaviorActionPropertyPage(PropertyPageBase):
    order = 15

    def __init__(self, item, event_manager):
        self.subject = item
        self.event_manager = event_manager

    def construct(self):
        builder = new_builder("call-action-editor")

        dropdown = builder.get_object("behavior")
        options = self._behavior_options()
        dropdown.set_model(options)

        if self.subject.behavior:
            dropdown.set_selected(
                next(
                    n
                    for n, lv in enumerate(options)
                    if lv.value == self.subject.behavior.id
                )
            )

        dropdown.connect("notify::selected", self._on_behavior_changed)

        return builder.get_object("call-action-editor")

    def _behavior_options(self):
        options = Gio.ListStore.new(LabelValue)
        options.append(LabelValue("", None))

        for c in sorted(
            (c for c in self.subject.model.select(UML.Behavior) if c.name),
            key=lambda c: c.name or "",
        ):
            options.append(LabelValue(c.name, c.id))

        return options

    def _on_behavior_changed(self, dropdown, _pspec):
        with Transaction(self.event_manager):
            if id := dropdown.get_selected_item().value:
                element = self.subject.model.lookup(id)
                assert isinstance(element, UML.Behavior)
                self.subject.behavior = element
            else:
                del self.subject.behavior


@PropertyPages.register(DecisionNodeItem)
class DecisionNodePropertyPage(PropertyPageBase):
    order = 20

    def __init__(self, item, event_manager):
        self.item = item
        self.event_manager = event_manager

    def construct(self):
        builder = new_builder(
            "decision-node-editor",
            signals={"show-type-changed": (self._on_show_type_change,)},
        )

        show_type = builder.get_object("show-type")
        show_type.set_active(self.item.show_underlying_type)

        return builder.get_object("decision-node-editor")

    def _on_show_type_change(self, button, gparam):
        with Transaction(self.event_manager):
            self.item.show_underlying_type = button.get_active()


@PropertyPages.register(ForkNodeItem)
class ForkNodePropertyPage(PropertyPageBase):
    order = 20

    def __init__(self, item, event_manager):
        self.item = item
        self.event_manager = event_manager

    def construct(self):
        builder = new_builder(
            "fork-node-editor",
            signals={"horizontal-changed": (self._on_horizontal_change,)},
        )

        horizontal = builder.get_object("horizontal")
        horizontal.set_active(self.is_horizontal())

        return builder.get_object("fork-node-editor")

    def is_horizontal(self):
        return self.item.matrix[2] != 0

    def _on_horizontal_change(self, button, gparam):
        active = button.get_active()
        horizontal = self.is_horizontal()
        with Transaction(self.event_manager):
            if active and not horizontal:
                self.item.matrix.rotate(math.pi / 2)
            elif not active and horizontal:
                self.item.matrix.rotate(-math.pi / 2)
            self.item.request_update()


@PropertyPages.register(UML.JoinNode)
class JoinNodePropertyPage(PropertyPageBase):
    order = 15

    subject: UML.JoinNode

    def __init__(self, subject, event_manager):
        self.subject = subject
        self.event_manager = event_manager

    def construct(self):
        subject = self.subject

        if not subject:
            return

        builder = new_builder(
            "join-node-editor",
            signals={"join-spec-changed": (self._on_join_spec_change,)},
        )

        join_spec = builder.get_object("join-spec")
        join_spec.set_text(subject.joinSpec or "")

        return builder.get_object("join-node-editor")

    def _on_join_spec_change(self, entry):
        value = entry.get_text().strip()
        with Transaction(self.event_manager, context="editing"):
            if (
                self.subject.joinSpec is None
                or UML.recipes.get_literal_value_as_string(self.subject.joinSpec)
                != value
            ):
                joinSpec = self.subject.model.create(UML.LiteralString)
                joinSpec.value = value
                joinSpec.joinNode = self.subject
                self.subject.joinSpec = joinSpec


@PropertyPages.register(UML.ControlFlow)
@PropertyPages.register(UML.ObjectFlow)
class FlowPropertyPageAbstract(PropertyPageBase):
    """Flow item element editor."""

    order = 15

    subject: UML.ActivityEdge

    def __init__(self, subject, event_manager):
        self.subject = subject
        self.event_manager = event_manager
        self.watcher = subject and subject.watcher()

    def construct(self):
        subject = self.subject

        if not subject:
            return

        builder = new_builder(
            "transition-editor",
        )

        guard = builder.get_object("guard")
        guard.set_text(subject.guard or "")

        @handler_blocking(guard, "changed", self._on_guard_change)
        def handler(event):
            v = event.new_value
            if v != guard.get_text():
                guard.set_text(v or "")

        self.watcher.watch("guard", handler)

        return unsubscribe_all_on_destroy(
            builder.get_object("transition-editor"), self.watcher
        )

    def _on_guard_change(self, entry):
        value = entry.get_text().strip()
        with Transaction(self.event_manager, context="editing"):
            if (
                self.subject.guard is None
                or UML.recipes.get_literal_value_as_string(self.subject.guard) != value
            ):
                guard = self.subject.model.create(UML.LiteralString)
                guard.value = value
                guard.activityEdge = self.subject
                self.subject.guard = guard


@PropertyPages.register(UML.Pin)
class PinPropertyPage(PropertyPageBase):
    """Pin element editor."""

    order = 15

    subject: UML.Pin

    def __init__(self, subject, event_manager):
        self.subject = subject
        self.event_manager = event_manager
        self.watcher = subject and subject.watcher()

    def construct(self):
        subject = self.subject

        if not subject:
            return

        builder = new_builder(
            "pin-editor",
            signals={
                "multiplicity-lower-changed": (self._on_multiplicity_lower_change,),
                "multiplicity-upper-changed": (self._on_multiplicity_upper_change,),
            },
        )

        dropdown = builder.get_object("type")
        model = self.list_of_classifiers(subject.model)
        dropdown.set_model(model)

        if subject.type:
            dropdown.set_selected(
                next(n for n, lv in enumerate(model) if lv.value == subject.type.id)
            )

        dropdown.connect("notify::selected", self._on_type_changed)

        multiplicity_lower = builder.get_object("multiplicity-lower")
        multiplicity_lower.set_text(subject.lowerValue or "")

        multiplicity_upper = builder.get_object("multiplicity-upper")
        multiplicity_upper.set_text(subject.upperValue or "")

        return builder.get_object("pin-editor")

    def list_of_classifiers(self, element_factory):
        options = Gio.ListStore.new(LabelValue)
        options.append(LabelValue("", None))

        for c in sorted(
            (c for c in element_factory.select(UML.Classifier) if c.name),
            key=lambda c: c.name or "",
        ):
            options.append(LabelValue(c.name, c.id))

        return options

    def _on_type_changed(self, dropdown, _pspec):
        subject = self.subject
        with Transaction(self.event_manager):
            if id := dropdown.get_selected_item().value:
                element = subject.model.lookup(id)
                assert isinstance(element, UML.Type)
                subject.type = element
            else:
                del subject.type

    def _on_multiplicity_lower_change(self, entry):
        value = entry.get_text().strip()
        with Transaction(self.event_manager, context="editing"):
            self.subject.lowerValue = value

    def _on_multiplicity_upper_change(self, entry):
        value = entry.get_text().strip()
        with Transaction(self.event_manager, context="editing"):
            self.subject.upperValue = value



"""Defines a status window class for displaying the progress of a queue."""

from __future__ import annotations

from gi.repository import Adw, GLib, Gtk, Pango


class StatusWindow:
    """Create a borderless window on the parent, usually the main window, with
    a label and a progress bar.

    The progress bar is updated as the queue is updated.
    """

    def __init__(self, title: str, message: str, parent: Gtk.Widget | None = None):
        """Create the status window.

        The title parameter is the title of the window.  The message
        parameter is a string displayed near the progress bar to
        indicate what is happening.  The parent parameter is the parent
        window to display the window in.
        """
        vbox = Gtk.Box.new(Gtk.Orientation.VERTICAL, spacing=12)
        label = Gtk.Label.new(message)
        label.set_ellipsize(Pango.EllipsizeMode.MIDDLE)
        vbox.append(label)

        self.progress_bar = Gtk.ProgressBar.new()
        self.progress_bar.set_size_request(400, -1)
        vbox.append(self.progress_bar)

        self.window = Adw.Dialog.new()
        self.window.set_child(vbox)
        self.window.set_title(title)
        self.window.add_css_class("status-window")
        self.window.present(parent)

    def progress(self, percentage: float):
        """Update progress percentage (0..100)."""
        if self.progress_bar:
            self.progress_bar.set_fraction(min(percentage, 100.0) / 100.0)

    def done(self):
        """Close the status window."""
        if self.window:
            if GLib.main_depth() > 0:
                # If we close the status window to quick it will stay around
                # See https://gitlab.gnome.org/GNOME/libadwaita/-/issues/970
                window = self.window
                GLib.idle_add(window.close, priority=GLib.PRIORITY_LOW)
            else:
                self.window.close()
            self.window = None



from __future__ import annotations

import logging

from gi.repository import Gio, GObject, Gtk

from gaphor import UML
from gaphor.core import gettext
from gaphor.core.eventmanager import EventManager
from gaphor.core.format import format, parse
from gaphor.diagram.propertypages import (
    NamePropertyPage,
    PropertyPageBase,
    PropertyPages,
    help_link,
    new_resource_builder,
    unsubscribe_all_on_destroy,
)
from gaphor.transaction import Transaction
from gaphor.UML.classes.datatype import DataTypeItem
from gaphor.UML.classes.enumeration import EnumerationItem
from gaphor.UML.classes.interface import Folded, InterfaceItem
from gaphor.UML.classes.klass import ClassItem
from gaphor.UML.deployments.connector import ConnectorItem
from gaphor.UML.propertypages import (
    check_button_handlers,
    create_list_store,
    list_item_factory,
    list_view_activated,
    list_view_key_handler,
    text_field_handlers,
    update_list_store,
)

log = logging.getLogger(__name__)


new_builder = new_resource_builder("gaphor.UML.classes")


@PropertyPages.register(UML.NamedElement)
class NamedElementPropertyPage(NamePropertyPage):
    """An adapter which works for any named item view.

    It also sets up a table view which can be extended.
    """

    order = 10

    def __init__(self, subject: UML.NamedElement | None, event_manager):
        super().__init__(subject, event_manager)

    def construct(self):
        if (
            not self.subject
            or UML.recipes.is_metaclass(self.subject)
            or isinstance(
                self.subject, UML.ActivityPartition | UML.ActivityParameterNode
            )
        ):
            return

        return super().construct()


@PropertyPages.register(UML.Classifier)
class ClassifierPropertyPage(PropertyPageBase):
    order = 15

    def __init__(self, subject, event_manager):
        self.subject = subject
        self.event_manager = event_manager

    def construct(self):
        if UML.recipes.is_metaclass(self.subject):
            return

        builder = new_builder(
            "classifier-editor",
            signals={
                "abstract-changed": (self._on_abstract_change,),
                "is-final-specialization-changed": (
                    self._on_is_final_specialization_change,
                ),
            },
        )

        abstract = builder.get_object("abstract")
        abstract.set_active(self.subject.isAbstract)

        is_final_specialization = builder.get_object("is-final-specialization")
        is_final_specialization.set_active(self.subject.isFinalSpecialization)

        return builder.get_object("classifier-editor")

    def _on_abstract_change(self, button, gparam):
        with Transaction(self.event_manager):
            self.subject.isAbstract = button.get_active()

    def _on_is_final_specialization_change(self, button, gparam):
        with Transaction(self.event_manager):
            self.subject.isFinalSpecialization = button.get_active()


@PropertyPages.register(InterfaceItem)
class InterfacePropertyPage(PropertyPageBase):
    """Adapter which shows a property page for an interface view."""

    order = 15

    def __init__(self, item, event_manager):
        self.item = item
        self.event_manager = event_manager

    def construct(self):
        builder = new_builder(
            "interface-editor", signals={"folded-changed": (self._on_fold_change,)}
        )

        item = self.item

        connected_items = [
            c.item for c in item.diagram.connections.get_connections(connected=item)
        ]
        disallowed = (ConnectorItem,)
        can_fold = not any(isinstance(i, disallowed) for i in connected_items)

        folded = builder.get_object("folded")
        folded.set_active(item.folded != Folded.NONE)
        folded.set_sensitive(can_fold)

        return builder.get_object("interface-editor")

    def _on_fold_change(self, button, gparam):
        item = self.item

        fold = button.get_active()

        with Transaction(self.event_manager):
            item.folded = Folded.PROVIDED if fold else Folded.NONE


class AttributeView(GObject.Object):
    def __init__(
        self, attr: UML.Property | None, klass: UML.Class, event_manager: EventManager
    ):
        super().__init__()
        self.attr = attr
        self.klass = klass
        self.event_manager = event_manager

    @GObject.Property(type=str, default="")
    def attribute(self):
        return (format(self.attr, note=True) or " ") if self.attr else ""

    @attribute.setter  # type: ignore[no-redef]
    def attribute(self, value):
        with Transaction(self.event_manager):
            if not self.attr:
                if not value:
                    return

                model = self.klass.model
                self.attr = model.create(UML.Property)
                self.klass.ownedAttribute = self.attr
            parse(self.attr, value)

    @GObject.Property(type=bool, default=False)
    def static(self):
        return self.attr.isStatic if self.attr else False

    @static.setter  # type: ignore[no-redef]
    def static(self, value):
        if not self.attr:
            return

        with Transaction(self.event_manager):
            self.attr.isStatic = value

    @GObject.Property(type=bool, default=False)
    def read_only(self):
        return self.attr.isReadOnly if self.attr else False

    @read_only.setter  # type: ignore[no-redef]
    def read_only(self, value):
        if not self.attr:
            return

        with Transaction(self.event_manager):
            self.attr.isReadOnly = value

    editing = GObject.Property(type=bool, default=False)

    def empty(self):
        return not self.attr

    def unlink(self):
        if self.attr:
            with Transaction(self.event_manager):
                self.attr.unlink()

    def swap(self, item1, item2):
        return self.klass.ownedAttribute.swap(item1.attr, item2.attr)


def attribute_model(klass: UML.Class, event_manager: EventManager) -> Gio.ListStore:
    return create_list_store(
        AttributeView,
        (a for a in klass.ownedAttribute if not a.association),
        lambda attr: AttributeView(attr, klass, event_manager),
    )


def update_attribute_model(
    store: Gio.ListStore, klass: UML.Class, event_manager: EventManager
) -> None:
    update_list_store(
        store,
        lambda item: item.attr,
        (a for a in klass.ownedAttribute if not a.association),
        lambda attr: AttributeView(attr, klass, event_manager),
    )


# We need a type check here, or all Class subtypes will get this editor
ATTRIBUTES_PAGE_CLASSIFIERS = [
    UML.Class,
    UML.DataType,
    UML.Enumeration,
    UML.Interface,
    UML.PrimitiveType,
    UML.Stereotype,
]

OPERATIONS_PAGE_CLASSIFIERS = list(ATTRIBUTES_PAGE_CLASSIFIERS)


@PropertyPages.register(UML.DataType)
@PropertyPages.register(UML.Class)
@PropertyPages.register(UML.Interface)
class AttributesPage(PropertyPageBase):
    order = 20

    def __init__(self, subject, event_manager: EventManager):
        super().__init__()
        self.subject = subject
        self.event_manager = event_manager
        self.watcher = subject and subject.watcher()

    def construct(self):
        if type(self.subject) not in ATTRIBUTES_PAGE_CLASSIFIERS:
            return

        builder = new_builder(
            "attributes-editor",
            "attributes-info",
            signals={
                "attributes-activated": (list_view_activated,),
                "attributes-key-pressed": (list_view_key_handler,),
                "attributes-info-clicked": (self.on_attributes_info_clicked,),
            },
        )
        self.info = builder.get_object("attributes-info")
        help_link(builder, "attributes-info-icon", "attributes-info")

        column_view: Gtk.ColumnView = builder.get_object("attributes-list")

        for column, factory in zip(
            column_view.get_columns(),
            [
                list_item_factory(
                    "text-field-cell.ui",
                    klass=AttributeView,
                    attribute=AttributeView.attribute,
                    placeholder_text=gettext("New Attribute…"),
                    signal_handlers=text_field_handlers("attribute"),
                ),
                list_item_factory(
                    "check-button-cell.ui",
                    klass=AttributeView,
                    attribute=AttributeView.read_only,
                    signal_handlers=check_button_handlers("read_only"),
                ),
                list_item_factory(
                    "check-button-cell.ui",
                    klass=AttributeView,
                    attribute=AttributeView.static,
                    signal_handlers=check_button_handlers("static"),
                ),
            ],
            strict=False,
        ):
            column.set_factory(factory)

        self.model = attribute_model(self.subject, self.event_manager)
        selection = Gtk.SingleSelection.new(self.model)
        column_view.set_model(selection)

        if self.watcher:
            self.watcher.watch("ownedAttribute", self.on_attributes_changed)

        return unsubscribe_all_on_destroy(
            builder.get_object("attributes-editor"), self.watcher
        )

    def on_attributes_changed(self, event):
        update_attribute_model(self.model, self.subject, self.event_manager)

    def on_attributes_info_clicked(self, image, event):
        self.info.set_visible(True)


@PropertyPages.register(DataTypeItem)
@PropertyPages.register(ClassItem)
@PropertyPages.register(EnumerationItem)
@PropertyPages.register(InterfaceItem)
class ShowAttributesPage(PropertyPageBase):
    order = 21

    def __init__(self, item, event_manager: EventManager):
        super().__init__()
        self.item = item
        self.event_manager = event_manager

    def construct(self):
        if not self.item.subject:
            return

        builder = new_builder(
            "show-attributes-editor",
            signals={
                "show-attributes-changed": (self.on_show_attributes_changed,),
            },
        )

        show_attributes = builder.get_object("show-attributes")
        show_attributes.set_active(self.item.show_attributes)

        return builder.get_object("show-attributes-editor")

    def on_show_attributes_changed(self, button, gparam):
        with Transaction(self.event_manager):
            self.item.show_attributes = button.get_active()


class OperationView(GObject.Object):
    def __init__(
        self, oper: UML.Operation | None, klass: UML.Class, event_manager: EventManager
    ):
        super().__init__()
        self.oper = oper
        self.klass = klass
        self.event_manager = event_manager

    @GObject.Property(type=str, default="")
    def operation(self):
        return (format(self.oper, note=True) or " ") if self.oper else ""

    @operation.setter  # type: ignore[no-redef]
    def operation(self, value):
        with Transaction(self.event_manager):
            if not self.oper:
                if not value:
                    return

                model = self.klass.model
                self.oper = model.create(UML.Operation)
                self.klass.ownedOperation = self.oper
            parse(self.oper, value)

    @GObject.Property(type=bool, default=False)
    def static(self):
        return self.oper.isStatic if self.oper else False

    @static.setter  # type: ignore[no-redef]
    def static(self, value):
        if not self.oper:
            return

        with Transaction(self.event_manager):
            self.oper.isStatic = value

    @GObject.Property(type=bool, default=False)
    def abstract(self):
        return self.oper.isAbstract if self.oper else False

    @abstract.setter  # type: ignore[no-redef]
    def abstract(self, value):
        if not self.oper:
            return

        with Transaction(self.event_manager):
            self.oper.isAbstract = value

    editing = GObject.Property(type=bool, default=False)

    def empty(self):
        return not self.oper

    def unlink(self):
        if self.oper:
            with Transaction(self.event_manager):
                self.oper.unlink()

    def swap(self, item1, item2):
        return self.klass.ownedOperation.swap(item1.oper, item2.oper)


def operation_model(klass: UML.Class, event_manager: EventManager) -> Gio.ListStore:
    return create_list_store(
        OperationView,
        klass.ownedOperation,
        lambda oper: OperationView(oper, klass, event_manager),
    )


def update_operation_model(
    store: Gio.ListStore, klass: UML.Class, event_manager: EventManager
) -> None:
    update_list_store(
        store,
        lambda item: item.oper,
        klass.ownedOperation,
        lambda oper: OperationView(oper, klass, event_manager),
    )


@PropertyPages.register(UML.DataType)
@PropertyPages.register(UML.Class)
@PropertyPages.register(UML.Interface)
class OperationsPage(PropertyPageBase):
    order = 30

    def __init__(self, subject, event_manager: EventManager):
        super().__init__()
        self.subject = subject
        self.event_manager = event_manager
        self.watcher = subject and subject.watcher()

    def construct(self):
        if type(self.subject) not in OPERATIONS_PAGE_CLASSIFIERS:
            return

        builder = new_builder(
            "operations-editor",
            "operations-info",
            signals={
                "operations-activated": (list_view_activated,),
                "operations-key-pressed": (list_view_key_handler,),
                "operations-info-clicked": (self.on_operations_info_clicked,),
            },
        )

        self.info = builder.get_object("operations-info")
        help_link(builder, "operations-info-icon", "operations-info")

        column_view: Gtk.ColumnView = builder.get_object("operations-list")

        for column, factory in zip(
            column_view.get_columns(),
            [
                list_item_factory(
                    "text-field-cell.ui",
                    klass=OperationView,
                    attribute=OperationView.operation,
                    placeholder_text=gettext("New Operation…"),
                    signal_handlers=text_field_handlers("operation"),
                ),
                list_item_factory(
                    "check-button-cell.ui",
                    klass=OperationView,
                    attribute=OperationView.abstract,
                    signal_handlers=check_button_handlers("abstract"),
                ),
                list_item_factory(
                    "check-button-cell.ui",
                    klass=OperationView,
                    attribute=OperationView.static,
                    signal_handlers=check_button_handlers("static"),
                ),
            ],
            strict=False,
        ):
            column.set_factory(factory)

        self.model = operation_model(self.subject, self.event_manager)
        selection = Gtk.SingleSelection.new(self.model)
        column_view.set_model(selection)

        if self.watcher:
            self.watcher.watch("ownedOperation", self.on_operations_changed)

        return unsubscribe_all_on_destroy(
            builder.get_object("operations-editor"), self.watcher
        )

    def on_operations_changed(self, event):
        update_operation_model(self.model, self.subject, self.event_manager)

    def on_operations_info_clicked(self, image, event):
        self.info.set_visible(True)


@PropertyPages.register(DataTypeItem)
@PropertyPages.register(EnumerationItem)
@PropertyPages.register(ClassItem)
@PropertyPages.register(InterfaceItem)
class ShowOperationsPage(PropertyPageBase):
    order = 31

    def __init__(self, item, event_manager):
        super().__init__()
        self.item = item
        self.event_manager = event_manager

    def construct(self):
        if not self.item.subject:
            return

        builder = new_builder(
            "show-operations-editor",
            signals={
                "show-operations-changed": (self.on_show_operations_changed,),
            },
        )

        show_operations = builder.get_object("show-operations")
        show_operations.set_active(self.item.show_operations)

        return builder.get_object("show-operations-editor")

    def on_show_operations_changed(self, button, gparam):
        with Transaction(self.event_manager):
            self.item.show_operations = button.get_active()


@PropertyPages.register(UML.Component)
class ComponentPropertyPage(PropertyPageBase):
    order = 15

    subject: UML.Component

    def __init__(self, subject, event_manager):
        self.subject = subject
        self.event_manager = event_manager

    def construct(self):
        subject = self.subject

        if not subject:
            return

        builder = new_builder(
            "component-editor",
            signals={"indirectly-instantiated-changed": (self._on_ii_change,)},
        )

        ii = builder.get_object("indirectly-instantiated")
        ii.set_active(subject.isIndirectlyInstantiated)

        return builder.get_object("component-editor")

    def _on_ii_change(self, button, gparam):
        """Called when user clicks "Indirectly instantiated" check button."""
        if subject := self.subject:
            with Transaction(self.event_manager):
                subject.isIndirectlyInstantiated = button.get_active()



from __future__ import annotations

import sys
from itertools import chain

from gi.repository import Gio, GLib, Gtk

from gaphor.abc import ActionProvider
from gaphor.action import action
from gaphor.entrypoint import load_entry_points


def apply_application_actions(component_registry, gtk_app) -> None:
    scope = "app"
    for _name, provider in component_registry.all(ActionProvider):
        for attrname, act in iter_actions(type(provider), scope):
            a = create_gio_action(act, provider, attrname)
            gtk_app.add_action(a)


def apply_shortcuts_from_entry_point(
    entry_point, scope, gtk_app, *, extra=None
) -> None:
    for provider_class in chain(load_entry_points(entry_point).values(), extra or []):
        if not issubclass(provider_class, ActionProvider):
            continue
        for _attrname, act in iter_actions(provider_class, scope):
            if act.shortcut:
                gtk_app.set_accels_for_action(
                    f"{scope}.{act.name}",
                    [_platform_modifier(s) for s in act.shortcuts],
                )


def window_action_group(component_registry, scope="win") -> Gio.SimpleActionGroup:
    action_group = Gio.SimpleActionGroup.new()
    for _name, provider in component_registry.all(ActionProvider):
        for attrname, act in iter_actions(type(provider), scope):
            a = create_gio_action(act, provider, attrname)
            action_group.add_action(a)
    return action_group


def apply_action_group(provider, scope, target_widget):
    action_group, shortcuts = create_action_group_and_shortcuts(provider, scope)
    target_widget.insert_action_group(scope, action_group)
    target_widget.add_controller(shortcuts)


def create_action_group_and_shortcuts(
    provider, scope
) -> tuple[Gio.SimpleActionGroup, Gtk.ShortcutController]:
    action_group = Gio.SimpleActionGroup.new()
    store = Gio.ListStore.new(Gtk.Shortcut)
    for attrname, act in iter_actions(type(provider), scope):
        a = create_gio_action(act, provider, attrname)
        action_group.add_action(a)
        for shortcut in act.shortcuts:
            store.append(named_shortcut(shortcut, act.detailed_name))

    ctrl = Gtk.ShortcutController.new_for_model(store)
    ctrl.set_scope(Gtk.ShortcutScope.LOCAL)
    return (action_group, ctrl)


def named_shortcut(shortcut, detailed_name):
    return Gtk.Shortcut.new(
        trigger=Gtk.ShortcutTrigger.parse_string(_platform_modifier(shortcut)),
        action=Gtk.NamedAction.new(detailed_name),
    )


def _platform_modifier(shortcut):
    return shortcut.replace(
        "<Primary>", "<Meta>" if sys.platform == "darwin" else "<Control>"
    )


def create_gio_action(act, provider, attrname):
    if act.state is not None:
        state = act.state(provider) if callable(act.state) else act.state
        a = Gio.SimpleAction.new_stateful(
            act.name, as_variant_type(act.arg_type), to_variant(state)
        )
        a.connect("change-state", _action_activate, provider, attrname)
    else:
        a = Gio.SimpleAction.new(act.name, as_variant_type(act.arg_type))
        a.connect("activate", _action_activate, provider, attrname)
    return a


def iter_actions(provider_class, scope):
    for attrname in dir(provider_class):
        func = getattr(provider_class, attrname)
        if callable(func):
            act: action | None = getattr(func, "__action__", None)
            if act and act.scope == scope:
                yield (attrname, act)


_GVARIANT_TYPE_STR = GLib.VariantType.new("s")
_GVARIANT_TYPE_INT = GLib.VariantType.new("i")
_GVARIANT_TYPE_BOOL = GLib.VariantType.new("b")


def as_variant_type(t):
    if t is None:
        return None
    elif t is str:
        return _GVARIANT_TYPE_STR
    elif t is int:
        return _GVARIANT_TYPE_INT
    elif t is bool:
        return _GVARIANT_TYPE_BOOL
    else:
        raise ValueError(f"No GVariantType declared for Python type {t}")


def to_variant(v):
    if v is None:
        return None
    elif isinstance(v, str):
        return GLib.Variant.new_string(v)
    elif isinstance(v, bool):
        return GLib.Variant.new_boolean(v)
    elif isinstance(v, int):
        return GLib.Variant.new_int32(v)
    else:
        raise ValueError(
            f"No GVariant mapping declared for Python value {v} of type {type(v)}"
        )


def from_variant(v):
    if v is None:
        return None
    elif _GVARIANT_TYPE_STR.equal(v.get_type()):
        return v.get_string()
    elif _GVARIANT_TYPE_INT.equal(v.get_type()):
        return v.get_int32()
    elif _GVARIANT_TYPE_BOOL.equal(v.get_type()):
        return v.get_boolean()
    else:
        raise ValueError(f"No Python mapping declared for GVariant value {v}")


def _action_activate(action, param, obj, name):
    method = getattr(obj, name)
    if param is not None:
        method(from_variant(param))
        if action.get_state_type():
            action.set_state(param)
    else:
        method()



"""C4 Model Language entrypoint."""

from gaphor.abc import ModelingLanguage
from gaphor.core.modeling import coremodel


class CoreModelingLanguage(ModelingLanguage):
    @property
    def name(self) -> str:
        return ""

    @property
    def toolbox_definition(self):
        raise ValueError("No toolbox for the core model")

    @property
    def diagram_types(self):
        raise ValueError("No diagram types for the core model")

    @property
    def element_types(self):
        raise ValueError("No element types for the core model")

    @property
    def model_browser_model(self):
        raise ValueError("No model browser model for the core model")

    def lookup_element(self, name, ns=None):
        assert ns in ("Core", None)
        return getattr(coremodel, name, None)


class MockModelingLanguage(ModelingLanguage):
    """This class can be used to instantly combine modeling languages."""

    def __init__(self, *modeling_languages: ModelingLanguage):
        self._modeling_languages = modeling_languages

    @property
    def name(self) -> str:
        return "Mock"

    @property
    def toolbox_definition(self):
        raise ValueError("No toolbox for the mock model")

    @property
    def diagram_types(self):
        raise ValueError("No diagram types for the mock model")

    @property
    def element_types(self):
        raise ValueError("No element types for the mock model")

    @property
    def model_browser_model(self):
        raise ValueError("No model browser model for the mock model")

    def lookup_element(self, name, ns=None):
        if ns:
            # This is sort of hackish.
            # It's better to lookup modeling languages from entry points
            for m in self._modeling_languages:
                if m.__class__.__name__.lower().startswith(ns.lower()):
                    return m.lookup_element(name, ns)

            raise ValueError(
                f"Invalid namespace '{ns}', should be one of {self._modeling_languages}"
            )

        return next(
            filter(
                None,
                [
                    provider.lookup_element(name)
                    for provider in self._modeling_languages
                ],
            ),
            None,
        )



# ruff: noqa: F401

import gaphor.SysML.deletable
import gaphor.SysML.drop
import gaphor.SysML.modelinglanguage
import gaphor.SysML.painter
import gaphor.SysML.sysmlfmt

__modeling_language__ = "SysML"



from gaphas.geometry import Rectangle, distance_point_point_fast

from gaphor.core.format import format, parse
from gaphor.diagram.instanteditors import instant_editor, popup_entry, show_popover
from gaphor.transaction import Transaction
from gaphor.UML.classes.association import AssociationItem


@instant_editor.register(AssociationItem)
def association_item_editor(item, view, event_manager, pos=None) -> bool:
    """Text edit support for Named items."""

    subject = item.subject
    if not subject:
        return False

    end_item = None
    if pos and distance_point_point_fast(item.handles()[0].pos, pos) < 50:
        end_item = item.head_end
    elif pos and distance_point_point_fast(item.handles()[-1].pos, pos) < 50:
        end_item = item.tail_end

    if end_item:
        text = (
            format(
                end_item.subject,
                visibility=True,
                is_derived=True,
                type=False,
                multiplicity=True,
                default=True,
            )
            or ""
        )

        entry = popup_entry(text)

        def update_text():
            assert end_item
            with Transaction(event_manager, context="editing"):
                parse(end_item.subject, entry.get_buffer().get_text())

        bb = end_item.name_bounds
        x, y = view.get_matrix_i2v(item).transform_point(bb.x, bb.y)
        box = Rectangle(x, y, 10, 10)
    else:
        text = item.subject.name or ""

        entry = popup_entry(text)

        def update_text():
            with Transaction(event_manager, context="editing"):
                item.subject.name = entry.get_buffer().get_text()

        box = item.middle_shape_size
        i2v = view.get_matrix_i2v(item)
        x, y = i2v.transform_point(box.x, box.y)
        w, h = i2v.transform_distance(box.width, box.height)
        box = Rectangle(x, y, w, h)

    show_popover(entry, view, box, update_text)
    return True



"""For ease of creation, maintain a mapping from Base to Diagram Item."""

from gaphor.core.modeling import Base, Presentation


def represents(uml_element, **metadata):
    """A decorator to assign a default Base type to a diagram item."""

    def wrapper(presentation):
        set_diagram_item(uml_element, presentation, metadata)
        return presentation

    return wrapper


# Map elements to their (default) representation.
_element_to_item_map: dict[type[Base], type[Presentation]] = {}
_item_to_metadata_map: dict[type[Presentation], dict[str, object]] = {}


def get_diagram_item(element_cls: type[Base]) -> type[Presentation] | None:
    global _element_to_item_map
    return _element_to_item_map.get(element_cls)


def get_diagram_item_metadata(item_cls: type[Presentation]):
    global _item_to_metadata_map
    return _item_to_metadata_map.get(item_cls, {})


def get_model_element(item_cls):
    global _element_to_item_map
    elements = [
        element
        for element, presentation in _element_to_item_map.items()
        if item_cls is presentation
    ]
    return elements[0] if elements else None


def set_diagram_item(element, item, metadata):
    assert element not in _element_to_item_map
    _element_to_item_map[element] = item
    _item_to_metadata_map[item] = metadata



from gaphor.C4Model import c4model
from gaphor.diagram.presentation import LinePresentation, Named, text_name
from gaphor.diagram.shapes import Box, CssNode, Text, draw_arrow_head
from gaphor.diagram.support import represents
from gaphor.UML.compartments import text_stereotypes


@represents(
    c4model.Dependency,
    head=c4model.Dependency.supplier,
    tail=c4model.Dependency.client,
)
class C4DependencyItem(Named, LinePresentation[c4model.Dependency]):
    def __init__(self, diagram, id=None):
        super().__init__(
            diagram,
            id,
            shape_middle=Box(
                text_stereotypes(self),
                text_name(self),
                text_technology(self),
            ),
        )

        self._handles[0].pos = (30, 20)
        self._handles[1].pos = (0, 0)

        self.watch("subject")
        self.watch("subject[UML:NamedElement].name")
        self.watch("subject[Dependency].technology")
        self.watch("subject[UML:Element].appliedStereotype.classifier.name")

        self.draw_head = draw_arrow_head


def text_technology(item: C4DependencyItem):
    return CssNode(
        "technology",
        item.subject,
        Text(
            text=lambda: item.subject
            and item.subject.technology
            and f"[{item.subject.technology}]"
            or ""
        ),
    )



"""Use case extension relationship."""

from gaphor import UML
from gaphor.diagram.presentation import LinePresentation, Named, text_name
from gaphor.diagram.shapes import Box, draw_arrow_head
from gaphor.diagram.support import represents
from gaphor.UML.compartments import text_stereotypes


@represents(UML.Extend, head=UML.Extend.extendedCase, tail=UML.Extend.extension)
class ExtendItem(Named, LinePresentation):
    """Use case extension relationship."""

    def __init__(self, diagram, id=None):
        super().__init__(
            diagram,
            id,
            shape_middle=Box(
                text_stereotypes(self, lambda: [self.diagram.gettext("extend")]),
                text_name(self),
            ),
        )

        self._handles[0].pos = (30, 20)
        self._handles[1].pos = (0, 0)

        self.watch("subject[Element].appliedStereotype.classifier.name").watch(
            "subject[NamedElement].name"
        )
        self.draw_head = draw_arrow_head



from __future__ import annotations

from gi.repository import Gio, Gtk

from gaphor.core import event_handler
from gaphor.core.changeset.apply import applicable, apply_change
from gaphor.core.modeling import ModelReady, PendingChange
from gaphor.event import (
    TransactionBegin,
    TransactionCommit,
    TransactionRollback,
)
from gaphor.i18n import translated_ui_string
from gaphor.transaction import Transaction
from gaphor.ui.modelmerge.organize import Node, organize_changes


class ModelMerge:
    def __init__(self, event_manager, element_factory, modeling_language):
        self.element_factory = element_factory
        self.event_manager = event_manager
        self.modeling_language = modeling_language
        self.model = Gio.ListStore.new(Node.__gtype__)
        self.selection = None
        self.page = None
        self.tree_view = None
        self._tx_depth = 0

    @property
    def needs_merge(self):
        return bool(next(self.element_factory.select(PendingChange), None))

    def refresh_model(self):
        self.model.remove_all()

        for node in organize_changes(self.element_factory, self.modeling_language):
            self.model.append(node)

        if self.model and self.page:
            self.page.set_visible(True)

    def open(self, builder):
        tree_model = Gtk.TreeListModel.new(
            self.model,
            passthrough=False,
            autoexpand=False,
            create_func=lambda node, _: node.children,
            user_data=None,
        )

        self.selection = Gtk.SingleSelection.new(tree_model)

        factory = Gtk.SignalListItemFactory.new()
        factory.connect("setup", list_item_factory_setup, self.apply)

        self.page = builder.get_object("modelmerge-page")
        self.tree_view = builder.get_object("modelmerge")
        self.tree_view.set_model(self.selection)
        self.tree_view.set_factory(factory)

        resolve = builder.get_object("modelmerge-resolve")
        resolve.connect("clicked", self.on_resolve_merge)

        self.event_manager.subscribe(self.on_model_loaded)

        self.refresh_model()

    def close(self):
        if self.page:
            self.page.set_visible(False)
        self.event_manager.unsubscribe(self.on_model_loaded)
        self.event_manager.unsubscribe(self.on_model_updated)

    def apply(self, change_node: Node | None):
        def do_apply(node):
            for element in node.elements:
                if applicable(element, self.element_factory):
                    apply_change(element, self.element_factory, self.modeling_language)
            if node.children:
                for n in node.children:
                    do_apply(n)

        if change_node:
            with Transaction(self.event_manager):
                do_apply(change_node)

        for item in self.model:
            item.sync()

    def on_resolve_merge(self, _button):
        pending_changes = self.element_factory.lselect(PendingChange)
        self.model.remove_all()
        with Transaction(self.event_manager):
            for change in pending_changes:
                change.unlink()
        self.close()

    @event_handler(ModelReady)
    def on_model_loaded(self, event):
        self.event_manager.subscribe(self.on_model_updated)
        self.refresh_model()

    @event_handler(TransactionBegin, TransactionCommit, TransactionRollback)
    def on_model_updated(self, event):
        self._tx_depth += +1 if isinstance(event, TransactionBegin) else -1
        if self._tx_depth == 0:
            self.apply(None)


def list_item_factory_setup(_factory, list_item, on_apply):
    builder = Gtk.Builder()
    builder.set_current_object(list_item)
    builder.extend_with_template(
        list_item,
        type(list_item).__gtype__,
        translated_ui_string("gaphor.ui.modelmerge", "change.ui"),
        -1,
    )
    apply = builder.get_object("apply")

    def on_active(button, _gparam):
        node = list_item.get_item().get_item()
        if button.get_active() and button.get_sensitive() and not node.applied:
            on_apply(node)

    apply.connect("notify::active", on_active)

    # TODO: popup



"""The SysML Modeling Language module is the entrypoint for SysML related
assets."""

from collections.abc import Iterable

from gaphor.abc import ModelingLanguage
from gaphor.core import gettext
from gaphor.diagram.diagramtoolbox import (
    DiagramType,
    ElementCreateInfo,
    ToolboxDefinition,
)
from gaphor.diagram.support import represents
from gaphor.SysML import diagramitems, sysml
from gaphor.SysML.toolbox import (
    sysml_diagram_types,
    sysml_element_types,
    sysml_toolbox_actions,
)
from gaphor.UML.general.diagramitem import DiagramItem
from gaphor.UML.treemodel import TreeModel

for _type in (
    sysml.ActivityDiagram,
    sysml.BlockDefinitionDiagram,
    sysml.InternalBlockDiagram,
    sysml.PackageDiagram,
    sysml.RequirementDiagram,
    sysml.SequenceDiagram,
    sysml.StateMachineDiagram,
    sysml.UseCaseDiagram,
):
    represents(_type)(DiagramItem)


class SysMLModelingLanguage(ModelingLanguage):
    @property
    def name(self) -> str:
        return gettext("SysML")

    @property
    def toolbox_definition(self) -> ToolboxDefinition:
        return sysml_toolbox_actions

    @property
    def diagram_types(self) -> Iterable[DiagramType]:
        yield from sysml_diagram_types

    @property
    def element_types(self) -> Iterable[ElementCreateInfo]:
        yield from sysml_element_types

    @property
    def model_browser_model(self) -> type[TreeModel]:
        return TreeModel

    def lookup_element(self, name, ns=None):
        assert ns in ("SysML", None)
        return getattr(sysml, name, None) or getattr(diagramitems, name, None)



from gaphas.connector import Handle, Port

from gaphor import UML
from gaphor.diagram.connectors import Connector
from gaphor.UML.actions.action import ActionItem, ValueSpecificationActionItem
from gaphor.UML.actions.pin import InputPinItem, OutputPinItem, PinItem


@Connector.register(ActionItem, PinItem)
class ActionPinConnector:
    def __init__(
        self,
        action: ActionItem,
        pin: PinItem,
    ) -> None:
        assert action.diagram is pin.diagram
        self.action = action
        self.pin = pin

    def allow(self, handle: Handle, port: Port) -> bool:
        return bool(self.action.diagram) and self.action.diagram is self.pin.diagram

    def connect(self, handle: Handle, port: Port) -> bool:
        """Connect and reconnect at model level.

        Returns `True` if a connection is established.
        """
        pin = self.pin
        if not pin.subject:
            pin.subject = pin.model.create(
                UML.InputPin if isinstance(pin, InputPinItem) else UML.OutputPin
            )

        assert isinstance(pin.subject, UML.InputPin | UML.OutputPin)
        pin.subject.opaqueAction = self.action.subject

        # This raises the item in the item hierarchy
        pin.change_parent(self.action)

        return True

    def disconnect(self, handle: Handle) -> None:
        pin = self.pin
        if pin.subject and pin.diagram:
            del pin.subject
            pin.change_parent(None)


@Connector.register(ValueSpecificationActionItem, PinItem)
class ValueSpecificationActionPinConnector(ActionPinConnector):
    def allow(self, handle: Handle, port: Port) -> bool:
        return (
            super().allow(handle, port)
            and not len(self.action.subject.outputValue)
            and isinstance(self.pin, OutputPinItem)
            and not self.pin.subject
        )



from __future__ import annotations

import importlib.resources
from unicodedata import normalize

from gi.repository import Gio, GObject, Pango

import gaphor.UML.uml as UML
from gaphor.core import event_handler
from gaphor.core.format import format
from gaphor.core.modeling import (
    Base,
    DerivedAdded,
    DerivedDeleted,
    DerivedSet,
    Diagram,
    ElementCreated,
    ElementDeleted,
    ElementUpdated,
    ModelFlushed,
    ModelReady,
)
from gaphor.diagram.group import Root, RootType, owner, owns
from gaphor.diagram.iconname import icon_name
from gaphor.i18n import gettext


class TreeItem(GObject.Object):
    def __init__(self, element: Base | None):
        super().__init__()
        self.element = element
        self.sync()

    icon = GObject.Property(type=str)
    icon_visible = GObject.Property(type=bool, default=False)

    readonly_text = GObject.Property(type=str)
    attributes = GObject.Property(type=Pango.AttrList)
    editing = GObject.Property(type=bool, default=False)
    can_edit = GObject.Property(type=bool, default=True)

    @GObject.Property(type=str)
    def editable_text(self):
        return (
            (self.element.name or "")
            if isinstance(self.element, UML.NamedElement)
            else ""
        )

    @editable_text.setter  # type: ignore[no-redef]
    def editable_text(self, text):
        if isinstance(self.element, UML.NamedElement):
            self.element.name = text or ""

    def sync(self) -> None:
        if element := self.element:
            self.readonly_text = format(element) or gettext("<None>")
            self.notify("editable-text")
            self.icon = icon_name(element)
            self.icon_visible = bool(
                self.icon
                and not isinstance(
                    element, UML.Parameter | UML.Property | UML.Operation
                )
            )
            self.attributes = pango_attributes(element)

    def start_editing(self):
        self.editing = True


class RelationshipItem(TreeItem):
    def __init__(self, child_model):
        super().__init__(None)
        self.child_model = child_model
        self.readonly_text = gettext("<Relationships>")
        self.can_edit = False


class Branch:
    def __init__(self):
        self.elements = Gio.ListStore.new(TreeItem.__gtype__)
        self.relationships = Gio.ListStore.new(TreeItem.__gtype__)

    def append(self, element: Base):
        if isinstance(element, UML.Relationship):
            if self.relationships.get_n_items() == 0:
                self.elements.insert(0, RelationshipItem(self.relationships))
            self.relationships.append(TreeItem(element))
        else:
            self.elements.append(TreeItem(element))

    def remove(self, element):
        list_store = (
            self.relationships
            if isinstance(element, UML.Relationship)
            else self.elements
        )
        if (
            index := next(
                (i for i, ti in enumerate(list_store) if ti.element is element),
                None,
            )
        ) is not None:
            list_store.remove(index)

        # Clean up empty relationships node
        if list_store is self.relationships and self.relationships.get_n_items() == 0:
            for i, e in enumerate(self.elements):
                if isinstance(e, RelationshipItem):
                    self.elements.remove(i)
                    break

    def remove_all(self):
        self.relationships.remove_all()
        self.elements.remove_all()

    def changed(self, element: Base):
        list_store = (
            self.relationships
            if isinstance(element, UML.Relationship)
            else self.elements
        )
        if not (
            tree_item := next((ti for ti in list_store if ti.element is element), None)
        ):
            return
        found, index = list_store.find(tree_item)
        if found:
            list_store.items_changed(index, 1, 1)

    def __len__(self):
        return self.elements.get_n_items()

    def __getitem__(self, index):
        return self.elements.get_item(index)

    def __iter__(self):
        yield from self.elements
        yield from self.relationships


def tree_item_sort(a: TreeItem, b: TreeItem) -> int:
    if isinstance(a, RelationshipItem):
        return -1
    if isinstance(b, RelationshipItem):
        return 1
    na = normalize("NFC", a.readonly_text).casefold()
    nb = normalize("NFC", b.readonly_text).casefold()
    return (na > nb) - (na < nb)


class TreeModel:
    def __init__(self, event_manager, element_factory, on_select=None, on_sync=None):
        super().__init__()
        self.branches: dict[TreeItem | RootType, Branch] = {Root: Branch()}
        self._on_select = on_select
        self._on_sync = on_sync
        self.event_manager = event_manager
        self.element_factory = element_factory

        event_manager.subscribe(self.on_element_created)
        event_manager.subscribe(self.on_element_deleted)
        event_manager.subscribe(self.on_owner_changed)
        event_manager.subscribe(self.on_owned_element_changed)
        event_manager.subscribe(self.on_attribute_changed)
        event_manager.subscribe(self.on_model_ready)

        self.on_model_ready()

    def shutdown(self) -> None:
        self.event_manager.unsubscribe(self.on_element_created)
        self.event_manager.unsubscribe(self.on_element_deleted)
        self.event_manager.unsubscribe(self.on_owner_changed)
        self.event_manager.unsubscribe(self.on_owned_element_changed)
        self.event_manager.unsubscribe(self.on_attribute_changed)
        self.event_manager.unsubscribe(self.on_model_ready)

    @property
    def template(self) -> str:
        return (importlib.resources.files("gaphor.UML") / "treeitem.ui").read_text(
            encoding="utf-8"
        )

    @property
    def root(self) -> Gio.ListStore:
        return self.branches[Root].elements

    def sync(self, element):
        if owner(element) and (tree_item := self.tree_item_for_element(element)):
            tree_item.sync()
            if self._on_sync:
                self._on_sync()

    def child_model(self, item: TreeItem) -> Gio.ListStore | None:
        """This method will create branches on demand (lazy)."""
        branches = self.branches
        if item in branches:
            return branches[item].elements
        elif isinstance(item, RelationshipItem):
            return item.child_model
        elif not item.element:
            return None
        elif owned_elements := owns(item.element):
            new_branch = Branch()
            self.branches[item] = new_branch
            for e in owned_elements:
                new_branch.append(e)
            return new_branch.elements
        return None

    def tree_item_sort(self, a, b) -> int:
        return tree_item_sort(a, b)

    def should_expand(self, item: TreeItem, element: Base) -> bool:
        return isinstance(element, UML.Relationship) and isinstance(
            item, RelationshipItem
        )

    def owner_branch_for_element(self, element: Base) -> Branch | None:
        if (own := owner(element)) is Root:
            return self.branches[Root]

        return next(
            (
                m
                for ti, m in self.branches.items()
                if isinstance(ti, TreeItem) and ti.element is own
            ),
            None,
        )

    def tree_item_for_element(self, element: Base | RootType) -> TreeItem | None:
        if element is Root:
            return None
        if owner_branch := self.owner_branch_for_element(element):
            return next((ti for ti in owner_branch if ti.element is element), None)
        return None

    def add_element(self, element: Base) -> None:
        if (not owner(element)) or self.tree_item_for_element(element):
            return

        if (owner_branch := self.owner_branch_for_element(element)) is not None:
            owner_branch.append(element)
        elif isinstance((own := owner(element)), Base):
            self.notify_child_model(own)

    def remove_element(self, element: Base) -> None:
        if not isinstance(element, Base):
            return

        for child in owns(element):
            self.remove_element(child)

        for owner_branch in [
            b for b in self.branches.values() if element in (i.element for i in b)
        ]:
            owner_branch.remove(element)

            if not len(owner_branch):
                self.remove_branch(owner_branch)

    def remove_branch(self, branch: Branch) -> None:
        tree_item = next(ti for ti, b in self.branches.items() if b is branch)
        if tree_item is Root:
            # Do never remove the root branch
            return

        del self.branches[tree_item]

        if tree_item.element:
            self.notify_child_model(tree_item.element)

    def notify_child_model(self, element: Base):
        # Only notify the change, the branch is created in child_model()
        if not self.branches.get(self.tree_item_for_element(element) or Root) and (
            owner_branch := self.branches.get(
                self.tree_item_for_element(owner(element) or Root) or Root
            )
        ):
            owner_branch.changed(element)

    def clear(self) -> None:
        root = self.branches[Root]
        root.remove_all()
        self.branches.clear()
        self.branches[Root] = root

    @event_handler(ElementCreated)
    def on_element_created(self, event: ElementCreated):
        self.add_element(event.element)

    @event_handler(ElementDeleted)
    def on_element_deleted(self, event: ElementDeleted):
        self.remove_element(event.element)

    @event_handler(DerivedAdded, DerivedDeleted)
    def on_owned_element_changed(self, event):
        """Ensure we update the node once owned elements change."""
        if event.property in (UML.Element.ownedElement, UML.Namespace.member):
            self.notify_child_model(event.element)

    @event_handler(DerivedSet)
    def on_owner_changed(self, event: DerivedSet):
        # Should check on ownedElement as well, since it may not have been updated
        # before this thing triggers
        if (
            event.property in (UML.Element.owner, UML.NamedElement.memberNamespace)
        ) and owner(event.element):
            element = event.element
            self.remove_element(element)
            self.add_element(element)
            if self._on_select:
                self._on_select(element)

    @event_handler(ElementUpdated)
    def on_attribute_changed(self, event: ElementUpdated):
        self.sync(event.element)

    @event_handler(ModelReady, ModelFlushed)
    def on_model_ready(self, _event=None):
        self.clear()

        for element in self.element_factory.select(owner):
            self.add_element(element)


def pango_attributes(element):
    attrs = Pango.AttrList.new()
    attrs.insert(
        Pango.attr_weight_new(
            Pango.Weight.BOLD if isinstance(element, Diagram) else Pango.Weight.NORMAL
        )
    )
    attrs.insert(
        Pango.attr_style_new(
            Pango.Style.ITALIC
            if isinstance(element, UML.Classifier | UML.BehavioralFeature)
            and element.isAbstract
            else Pango.Style.NORMAL
        )
    )
    return attrs



"""Save Gaphor models to Gaphors own XML format."""

__all__ = ["save"]

import logging
from collections.abc import Callable, Iterable
from functools import partial

from gaphor import application
from gaphor.core.modeling import Base, ElementFactory
from gaphor.core.modeling.collection import collection
from gaphor.storage.xmlwriter import WriterProtocol, XMLWriter

FILE_FORMAT_VERSION = "4"
MODEL_NS = "https://gaphor.org/model"
MODELING_LANGUAGE_NS = "https://gaphor.org/modelinglanguage"

log = logging.getLogger(__name__)


def save(
    out: WriterProtocol,
    element_factory: ElementFactory,
    status_queue: Callable[[float], None] | None = None,
):
    for status in save_generator(out, element_factory):
        if status_queue:
            status_queue(status)


def save_generator(
    out: WriterProtocol, element_factory: ElementFactory
) -> Iterable[float]:
    """Save the current model using @writer, which is a
    gaphor.storage.xmlwriter.XMLWriter instance."""

    with XMLWriter(out).document() as writer:
        writer.prefix_mapping("", MODEL_NS)
        for ml in sorted({e.__modeling_language__ for e in element_factory}):
            writer.prefix_mapping(ml, f"{MODELING_LANGUAGE_NS}/{ml}")

        with writer.element_ns(
            (MODEL_NS, "gaphor"),
            {
                (MODEL_NS, "version"): FILE_FORMAT_VERSION,
                (MODEL_NS, "gaphor-version"): application.distribution().version,
            },
        ):
            with writer.element_ns((MODEL_NS, "model"), {}):
                size = element_factory.size()
                save_func = partial(
                    save_element, element_factory=element_factory, writer=writer
                )
                for n, e in enumerate(element_factory, start=1):
                    clazz = e.__class__.__name__
                    assert e.id
                    ns = f"{MODELING_LANGUAGE_NS}/{e.__modeling_language__}"
                    with writer.element_ns((ns, clazz), {(MODEL_NS, "id"): str(e.id)}):
                        e.save(save_func)

                    if n % 25 == 0:
                        yield (n * 100) / size


def save_element(
    name: str,
    value: str | int | bool | Base | collection[Base],
    element_factory: ElementFactory,
    writer: XMLWriter,
) -> None:
    """Save attributes and references from items in the gaphor.UML module.

    A value may be a primitive (string, int), a
    gaphor.core.modeling.collection (which contains a list of references
    to other UML elements) or a Diagram (which contains diagram items).
    """

    def resolvable(value):
        if value.id and value in element_factory:
            return True
        log.warning(
            f"Model has unknown reference {value.id}. Reference will be skipped."
        )
        return False

    def save_reference(name: str, value: Base):
        """Save a value as a reference to another element in the model.

        This applies to both UML and canvas items.
        """
        if resolvable(value):
            with writer.element(name, {}):
                with writer.element("ref", {"refid": value.id}):
                    pass

    def save_collection(name: str, value: collection[Base]):
        """Save a list of references."""
        if value:
            with writer.element(name, {}):
                with writer.element("reflist", {}):
                    for v in value:
                        if resolvable(v):
                            with writer.element("ref", {"refid": v.id}):
                                pass

    def save_value(name: str, value: str | int | bool):
        """Save a value (attribute)."""
        if value is not None:
            with writer.element(name, {}):
                with writer.element("val", {}):
                    writer.characters(str(value))

    if isinstance(value, Base):
        save_reference(name, value)
    elif isinstance(value, collection):
        save_collection(name, value)
    else:
        save_value(name, value)



from __future__ import annotations

from gaphas.geometry import Rectangle

from gaphor.core.modeling.diagram import DrawContext, StyledDiagram
from gaphor.diagram.painter import DiagramTypePainter
from gaphor.diagram.shapes import Box, CssNode, Orientation, Text, cairo_state, stroke
from gaphor.UML.recipes import get_applied_stereotypes, stereotypes_str
from gaphor.UML.uml import Diagram


@DiagramTypePainter.register(Diagram)  # type: ignore[attr-defined]
class UMLDiagramTypePainter:
    def __init__(self, diagram: Diagram):
        self.diagram = diagram
        self._pentagon = CssNode(
            "pentagon",
            diagram,
            Box(
                CssNode("diagramtype", None, Text(text=lambda: diagram.diagramType)),
                CssNode(
                    "stereotypes", None, Text(text=lambda: stereotypes_str(diagram))
                ),
                CssNode("name", None, Text(text=lambda: self.label())),
                orientation=Orientation.HORIZONTAL,
                draw=draw_pentagon,
            ),
        )

    def label(self):
        return self.diagram.name

    def paint(self, _items, cr):
        diagram = self.diagram
        if not diagram.diagramType and not any(get_applied_stereotypes(diagram)):
            return

        style = diagram.style(StyledDiagram(diagram))

        context = DrawContext(
            cairo=cr,
            style=style,
            selected=False,
            focused=False,
            hovered=False,
            dropzone=False,
        )

        with cairo_state(cr):
            cr.identity_matrix()

            size = self._pentagon.size(context)  # type: ignore[arg-type]
            bb = Rectangle(0, 0, *size)
            self._pentagon.draw(context, bb)


def draw_pentagon(box, context, bounding_box):
    cr = context.cairo
    w = bounding_box.width
    h = bounding_box.height
    line_width = context.style.get("line-width", 2)
    with cairo_state(cr):
        h2 = h / 2.0
        cr.move_to(-line_width, -line_width)
        cr.line_to(-line_width, h)
        cr.line_to(w - 4, h)
        cr.line_to(w, h2)
        cr.line_to(w, -line_width)
        stroke(context, fill=True)



"""Parser for CSS selectors, based on cssselect2.

Changes compared to cssselect2:

* The parser has been adapted to support statements like `class[subject.ownedAttribute=foo]`.
* :not is treated as a normal pseudo-class. It's a `FunctionalPseudoClassSelector`.

Original module: cssselect2.parser
:copyright: (c) 2012 by Simon Sapin, 2017 by Guillaume Ayoub.
:license: BSD, see https://github.com/Kozea/cssselect2/blob/master/LICENSE for more details.
"""

import tinycss2


def selectors(input, namespaces=None):
    """Parse tokens or string into selectors."""
    if isinstance(input, str):
        input = tinycss2.tokenizer.parse_component_value_list(input, skip_comments=True)
    tokens = TokenStream(input)
    namespaces = namespaces or {}
    yield parse_selector(tokens, namespaces)
    tokens.skip_whitespace_and_comment()
    while 1:
        next = tokens.next()
        if next is None:
            return
        elif next == ",":
            yield parse_selector(tokens, namespaces)
        else:
            raise SelectorError(next, f"unexpected {next.type} token.")


def media_query_selector(input):
    tokens = TokenStream(input)
    query = []
    while 1:
        tokens.skip_whitespace_and_comment()
        next = tokens.next()
        if next is None:
            break
        elif next.type in ("ident", "literal", "string"):
            query.append(next.value)
        elif next.type == "() block":
            # lazy: only parse inside the parenthesis
            # we do not support combinations at this time
            tokens = TokenStream(next.content)
        else:
            raise SelectorError(next, f"unexpected {next.type} token.")

    return MediaSelector(query)


def parse_selector(tokens, namespaces):
    result = parse_compound_selector(tokens, namespaces)
    while 1:
        has_whitespace = tokens.skip_whitespace()
        while tokens.skip_comment():
            has_whitespace = tokens.skip_whitespace() or has_whitespace
        peek = tokens.peek()
        if peek in (">", "+", "~"):
            combinator = peek.value
            tokens.next()
        elif peek is None or peek == "," or not has_whitespace:
            return result
        else:
            combinator = " "
        compound = parse_compound_selector(tokens, namespaces)
        result = CombinedSelector(result, combinator, compound)


def parse_compound_selector(tokens, namespaces):
    type_selectors = parse_type_selector(tokens, namespaces)
    simple_selectors = type_selectors if type_selectors is not None else []
    while 1:
        simple_selector = parse_simple_selector(tokens, namespaces)
        if simple_selector is None:
            break
        simple_selectors.append(simple_selector)

    if simple_selectors or type_selectors is not None:
        return CompoundSelector(simple_selectors)
    peek = tokens.peek()
    raise SelectorError(
        peek, f"expected a compound selector, got {peek.type if peek else 'EOF'}"
    )


def parse_type_selector(tokens, namespaces):
    tokens.skip_whitespace()
    qualified_name = parse_qualified_name(tokens, namespaces)
    if qualified_name is None:
        return None

    simple_selectors = []
    namespace, local_name = qualified_name
    if local_name is not None:
        simple_selectors.append(LocalNameSelector(local_name))
    if namespace is not None:
        simple_selectors.append(NamespaceSelector(namespace))  # type: ignore[arg-type]
    return simple_selectors


def parse_simple_selector(tokens, namespaces):
    peek = tokens.peek()
    if peek is None:
        return None
    if peek.type == "hash" and peek.is_identifier:
        tokens.next()
        return IDSelector(peek.value)
    elif peek == ".":
        tokens.next()
        next = tokens.next()
        if next is None or next.type != "ident":
            raise SelectorError(next, f"Expected a class name, got {next}")
        return ClassSelector(next.value)
    elif peek.type == "[] block":
        tokens.next()
        return parse_attribute_selector(TokenStream(peek.content), namespaces)
    elif peek == ":":
        tokens.next()
        next = tokens.next()
        if next == ":":
            next = tokens.next()
            if next is None or next.type != "ident":
                raise SelectorError(next, f"Expected a pseudo-element name, got {next}")
            return PseudoElementSelector(next.lower_value)
        elif next is not None and next.type == "ident":
            return PseudoClassSelector(next.lower_value)
        elif next is not None and next.type == "function":
            return FunctionalPseudoClassSelector(next.lower_name, next.arguments)
        else:
            raise SelectorError(next, f"unexpected {next} token.")
    else:
        return None


def parse_attribute_selector(tokens, namespaces):
    tokens.skip_whitespace()
    qualified_name = parse_qualified_name(tokens, namespaces, is_attribute=True)
    if qualified_name is None:
        next = tokens.next()
        raise SelectorError(next, f"expected attribute name, got {next}")
    namespace, local_name = qualified_name

    # Allow syntax like "subject.ownedAttribute"
    if local_name:
        name, lower_name = local_name
        while 1:
            peek = tokens.peek()
            if peek == ".":
                next = tokens.next()
                name += next.value
                lower_name += next.value
            elif peek and peek.type == "ident":
                next = tokens.next()
                name += next.value
                lower_name += next.lower_value
            else:
                break

        local_name = name, lower_name

    tokens.skip_whitespace()
    peek = tokens.peek()
    if peek is None:
        operator = None
        value = None
    elif peek in ("=", "~=", "|=", "^=", "$=", "*="):
        operator = peek.value
        tokens.next()
        tokens.skip_whitespace()
        next = tokens.next()
        if next is None or next.type not in ("ident", "string"):
            next_type = "None" if next is None else next.type
            raise SelectorError(next, f"expected attribute value, got {next_type}")
        value = next.value
    else:
        raise SelectorError(peek, f"expected attribute selector operator, got {peek}")

    tokens.skip_whitespace()
    next = tokens.next()
    if next is not None:
        raise SelectorError(next, f"expected ], got {next.type}")
    return AttributeSelector(namespace, local_name, operator, value)


def parse_qualified_name(tokens, namespaces, is_attribute=False):
    """Return ``(namespace, local)`` for given tokens.

    Can also return ``None`` for a wildcard.

    The empty string for ``namespace`` means "no namespace".
    """
    peek = tokens.peek()
    if peek is None:
        return None
    if peek.type == "ident":
        first_ident = tokens.next()
        peek = tokens.peek()
        if peek != "|":
            namespace = "" if is_attribute else namespaces.get(None, None)
            return namespace, (first_ident.value, first_ident.lower_value)
        tokens.next()
        namespace = namespaces.get(first_ident.value)
        if namespace is None:
            raise SelectorError(
                first_ident, f"undefined namespace prefix: {first_ident.value}"
            )

    elif peek == "*":
        next = tokens.next()
        peek = tokens.peek()
        if peek != "|":
            if is_attribute:
                raise SelectorError(next, f"Expected local name, got {next.type}")
            return namespaces.get(None, None), None
        tokens.next()
        namespace = None
    elif peek == "|":
        tokens.next()
        namespace = ""
    else:
        return None

    # If we get here, we just consumed '|' and set ``namespace``
    next = tokens.next()
    if next.type == "ident":
        return namespace, (next.value, next.lower_value)
    elif next == "*" and not is_attribute:
        return namespace, None
    else:
        raise SelectorError(next, f"Expected local name, got {next.type}")


class SelectorError(ValueError):
    """A specialized ``ValueError`` for invalid selectors."""


class TokenStream:
    def __init__(self, tokens):
        self.tokens = iter(tokens)
        self.peeked = []  # In reversed order

    def next(self):
        return self.peeked.pop() if self.peeked else next(self.tokens, None)

    def peek(self):
        if not self.peeked:
            self.peeked.append(next(self.tokens, None))
        return self.peeked[-1]

    def skip(self, skip_types):
        found = False
        while 1:
            peek = self.peek()
            if peek is None or peek.type not in skip_types:
                break
            self.next()
            found = True
        return found

    def skip_whitespace(self):
        return self.skip(["whitespace"])

    def skip_comment(self):
        return self.skip(["comment"])

    def skip_whitespace_and_comment(self):
        return self.skip(["comment", "whitespace"])


class MediaSelector:
    def __init__(self, query):
        self.query = query

    @property
    def specificity(self):
        return 0, 0, 0

    def __repr__(self):
        return "".join(self.query)


class CombinedSelector:
    def __init__(self, left, combinator, right):
        #: Combined or compound selector
        self.left = left
        # One of `` `` (a single space), ``>``, ``+`` or ``~``.
        self.combinator = combinator
        #: compound selector
        self.right = right

    @property
    def specificity(self):
        a1, b1, c1 = self.left.specificity
        a2, b2, c2 = self.right.specificity
        return a1 + a2, b1 + b2, c1 + c2

    def __repr__(self):
        return f"{self.left!r}{self.combinator}{self.right!r}"


class CompoundSelector:
    """Aka.

    sequence of simple selectors, in Level 3.
    """

    def __init__(self, simple_selectors):
        self.simple_selectors = simple_selectors

    @property
    def specificity(self):
        if self.simple_selectors:
            return tuple(
                map(
                    sum,
                    zip(
                        *(sel.specificity for sel in self.simple_selectors),
                        strict=False,
                    ),
                )
            )
        return 0, 0, 0

    def __repr__(self):
        return "".join(map(repr, self.simple_selectors))


class LocalNameSelector:
    specificity = 0, 0, 1

    def __init__(self, local_name):
        self.local_name, self.lower_local_name = local_name

    def __repr__(self):
        return self.local_name


class NamespaceSelector:
    specificity = 0, 0, 0

    def __init__(self, namespace):
        #: The namespace URL as a string,
        #: or the empty string for elements not in any namespace.
        self.namespace = namespace

    def __repr__(self):
        return "|" if self.namespace == "" else f"{{{self.namespace}}}|"


class IDSelector:
    specificity = 1, 0, 0

    def __init__(self, ident):
        self.ident = ident

    def __repr__(self):
        return f"#{self.ident}"


class ClassSelector:
    specificity = 0, 1, 0

    def __init__(self, class_name):
        self.class_name = class_name

    def __repr__(self):
        return f".{self.class_name}"


class AttributeSelector:
    specificity = 0, 1, 0

    def __init__(self, namespace, name, operator, value):
        self.namespace = namespace
        self.name, self.lower_name = name
        #: A string like ``=`` or ``~=``, or None for ``[attr]`` selectors
        self.operator = operator
        #: A string, or None for ``[attr]`` selectors
        self.value = value

    def __repr__(self):
        namespace = "*|" if self.namespace is None else f"{{{self.namespace}}}"
        return f"[{namespace}{self.name}{self.operator}{self.value!r}]"


class PseudoClassSelector:
    specificity = 0, 1, 0

    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return f":{self.name}"


class PseudoElementSelector:
    specificity = 0, 0, 1

    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return f"::{self.name}"


class FunctionalPseudoClassSelector:
    specificity = 0, 1, 0

    def __init__(self, name, arguments):
        self.name = name
        self.arguments = arguments

    def __repr__(self):
        return f":{self.name}{tuple(self.arguments)!r}"



from __future__ import annotations

import logging
from collections.abc import Iterable

from gi.repository import Gtk

from gaphor.core.eventmanager import EventManager
from gaphor.transaction import Transaction

log = logging.getLogger(__name__)


class TxData:
    def __init__(self, event_manager):
        self.event_manager = event_manager
        self.txs: list[Transaction] = []

    def begin(self):
        self.txs.append(Transaction(self.event_manager))

    def commit(self):
        assert self.txs
        tx = self.txs.pop()
        tx.commit()


def transactional_tool(
    *tools: Gtk.Gesture, event_manager: EventManager | None = None
) -> Iterable[Gtk.EventController]:
    tx_data = TxData(event_manager)
    for tool in tools:
        tool.connect("begin", on_begin, tx_data)
        tool.connect_after("end", on_end, tx_data)
        tool.connect_after("cancel", on_cancel, tx_data)

    key_ctrl = Gtk.EventControllerKey.new()
    key_ctrl.connect("key-pressed", key_pressed, tools, tx_data)
    return (*tools, key_ctrl)


def on_begin(gesture, _sequence, tx_data):
    while tx_data.txs:
        log.warning(
            "Closing transaction. This should have happened in the 'end' handler."
        )
        tx_data.commit()

    tx_data.begin()


def on_cancel(gesture, _sequence, tx_data):
    if tx_data.txs:
        Transaction.mark_rollback()


def on_end(gesture, _sequence, tx_data):
    tx_data.commit()


def key_pressed(ctrl, keyval, keycode, state, tools, tx_data):
    if tx_data.txs:
        for tool in tools:
            tool.reset()
        return True



#!/usr/bin/env python
"""Base class for model elements."""

from __future__ import annotations

import importlib
import logging
from collections.abc import Callable, Iterator
from dataclasses import dataclass
from functools import cache
from typing import TYPE_CHECKING, Protocol, TypeVar, overload
from uuid import uuid1

from gaphor.core.modeling.collection import collection
from gaphor.core.modeling.event import ElementTypeUpdated, ElementUpdated
from gaphor.core.modeling.properties import relation_many, umlproperty

if TYPE_CHECKING:
    from gaphor.core.modeling.diagram import Diagram
    from gaphor.core.modeling.presentation import Presentation

log = logging.getLogger(__name__)


@dataclass
class UnlinkEvent:
    """Used to tell event handlers this element should be unlinked.

    This event is used internally and should not be handled outside
    `gaphor.core.modeling`.
    """

    element: Base
    diagram: Diagram | None = None


Id = str


def uuid_generator() -> Iterator[Id]:
    while True:
        yield str(uuid1())


_generator: Iterator[Id] = uuid_generator()


def generate_id(generator: Iterator[Id] | None = None) -> Id:
    global _generator
    if generator:
        _generator = generator
    return next(_generator)


class classproperty:
    def __init__(self, func):
        self.fget = func

    def __get__(self, _instance, owner):
        return self.fget(owner)


class Base:
    """Base class for all model data classes."""

    presentation: relation_many[Presentation]

    def __init__(self, id: Id | None = None, model: RepositoryProtocol | None = None):
        """Create an element. As optional parameters an id and model can be
        given.

        Id is a serial number for the element. If no id is provided, one will automatically
        be created. Id's should be unique within the model.

        A model can be provided to refer to the model
        (:class:~gaphor.core.modeling.ElementFactory) this element belongs to.
        """
        self._id: Id = id or generate_id()
        # The model this element belongs to.
        # NOTE: Will be unset by ElementFactory once it's `unlink()`ed.
        self._model = model

    @property
    def id(self) -> Id:
        """An id (read-only), unique within the model."""
        return self._id

    @property
    def model(self) -> RepositoryProtocol:
        """The owning model, raises :class:`TypeError` when model is not set."""
        if not self._model:
            raise TypeError(
                "Can't retrieve the model since it's not set on construction"
            )
        return self._model

    @classproperty
    def __modeling_language__(cls) -> str | None:
        """The modeling language this class belongs to."""
        if ml := _resolve_modeling_language(cls.__module__):
            return ml
        raise AttributeError(
            f"module '{cls.__module__}' or its parents have no attribute '__modeling_language__'"
        )

    @classproperty
    def __properties__(cls) -> Iterator[umlproperty]:
        """Iterate over all attributes, associations, etc."""
        umlprop = umlproperty
        for propname in dir(cls):
            if not propname.startswith("_"):
                prop = getattr(cls, propname)
                if isinstance(prop, umlprop):
                    yield prop

    def __str__(self) -> str:
        return f"<{self.__class__.__module__}.{self.__class__.__name__} element {self._id}>"

    __repr__ = __str__

    def __setattr__(self, key, value) -> None:
        if key.startswith("_") or hasattr(self.__class__, key):
            super().__setattr__(key, value)
        else:
            raise AttributeError(
                f"Property {self.__class__.__name__}.{key} does not exist"
            )

    def save(
        self,
        save_func: Callable[[str, str | bool | int | Base | collection[Base]], None],
    ) -> None:
        """Save the state by calling ``save_func(name, value)``."""
        for prop in self.__properties__:
            prop.save(self, save_func)

    def load(
        self, name: str, value: str | bool | int | Base | collection[Base]
    ) -> None:
        """Loads value in name.

        Make sure that after all elements are loaded, postload() should be called.
        """
        prop = getattr(type(self), name)
        prop.load(self, value)

    def postload(self) -> None:
        """Fix up the odds and ends.

        This is run after all elements are loaded.
        """
        for prop in self.__properties__:
            prop.postload(self)

    def unlink(self) -> None:
        """Unlink the element.

        All the elements references are destroyed.
        For composite associations, the associated elements are also unlinked.

        The unlink lock is acquired while unlinking this element's
        properties to avoid recursion problems.
        """
        for prop in self.__properties__:
            prop.unlink(self)

        log.debug("unlinking %s", self)
        self.handle(UnlinkEvent(self))

    def handle(self, event: object) -> None:
        """Propagate incoming events.

        This only works if the element has been created by an :class:`~gaphor.core.modeling.ElementFactory`
        """
        if model := self._model:
            model.handle(event)

    def watcher(self, default_handler: Handler | None = None) -> EventWatcherProtocol:
        """Create a new watcher for this element.

        Watchers provide a convenient way to get signalled when a property relative to
        ``self`` has been changed.

        To use a watcher, the element should be created by a properly wired up :class:`~gaphor.core.modeling.ElementFactory``.

        This example is purely illustrative:

        >>> element = Base()
        >>> watcher = element.watcher(default_handler=print)
        >>> watcher.watch("note")  # Watch for changed on element.note
        """
        if model := self._model:
            return model.watcher(self, default_handler)
        return DummyEventWatcher()

    def isKindOf(self, class_: type[Base]) -> bool:
        """Returns :const:`True` if the object is an instance of ``class_``."""
        return isinstance(self, class_)

    def isTypeOf(self, other: Base) -> bool:
        """Returns :const:`True` if the object is of the same type as the ``other``."""
        return isinstance(self, type(other))


class DummyEventWatcher:
    def watch(self, path: str, handler: Handler | None = None) -> DummyEventWatcher:
        return self

    def unsubscribe_all(self, *_args) -> None:
        pass


T = TypeVar("T", bound=Base)

Handler = Callable[[ElementUpdated], None]


class RepositoryProtocol(Protocol):
    def create(self, type: type[T]) -> T: ...

    def create_as(self, type: type[T], id: str) -> T: ...

    @overload
    def select(self, expression: Callable[[Base], bool]) -> Iterator[Base]: ...

    @overload
    def select(self, type_: type[T]) -> Iterator[T]: ...

    @overload
    def select(self, expression: None) -> Iterator[Base]: ...

    def lookup(self, id: str) -> Base | None: ...

    def watcher(
        self, element: Base, default_handler: Handler | None = None
    ) -> EventWatcherProtocol: ...

    def handle(self, event: object) -> None: ...


class EventWatcherProtocol(Protocol):
    def watch(
        self, path: str, handler: Handler | None = None
    ) -> EventWatcherProtocol: ...

    def unsubscribe_all(self) -> None: ...


def swap_element_type(element: Base, new_class: type[Base]) -> None:
    """A "trick" to swap the element type.

    Used in certain cases where the underlying element type may change.
    """
    if element.__class__ is not new_class:
        old_class = element.__class__
        element.__class__ = new_class
        element.handle(ElementTypeUpdated(element, old_class))


@cache
def _resolve_modeling_language(module_name: str) -> str | None:
    mod = importlib.import_module(module_name)
    if ml := getattr(mod, "__modeling_language__", None):
        return str(ml)
    parent_name = module_name.rsplit(".", 1)[0]
    if parent_name == module_name:
        return None
    return _resolve_modeling_language(parent_name)



"""Activity Partition item."""

from gaphas.geometry import Rectangle
from gaphas.item import NW, SE
from gaphas.types import Pos

from gaphor import UML
from gaphor.core.modeling import DrawContext
from gaphor.core.modeling.properties import association, relation_many
from gaphor.diagram.presentation import ElementPresentation
from gaphor.diagram.shapes import DEFAULT_PADDING, Box, CssNode, Orientation, stroke
from gaphor.diagram.support import represents
from gaphor.diagram.text import Layout

HEADER_HEIGHT: int = 29


@represents(UML.ActivityPartition)
class PartitionItem(ElementPresentation[UML.ActivityPartition]):
    def __init__(self, diagram, id=None):
        super().__init__(diagram, id)
        self.min_height = 300
        self._loading = False
        self.watch("subject[NamedElement].name")
        self.watch("subject[Element].appliedStereotype.classifier.name")
        self.watch("partition", self.update_shapes)
        self.watch("partition.name")
        self.watch("partition[ActivityPartition].represents[NamedElement].name")
        self.handles()[NW].pos.add_handler(self.update_width)
        self.handles()[SE].pos.add_handler(self.update_width)

    def update_width(self, pos, oldpos) -> None:
        if self._loading:
            return

        if pos is self.handles()[SE].pos:
            x_w = self.handles()[NW].pos.x
            old_width = oldpos[0] - x_w
            new_width = pos.x - x_w
            left = x_w
            offset = 0
        else:
            x_e = self.handles()[SE].pos.x
            old_width = x_e - oldpos[0]
            new_width = x_e - pos.x
            left = oldpos[0]
            offset = pos.x - oldpos[0]

        if new_width < 1e-6:
            return
        factor = 1 - old_width / new_width
        for child in self.children:
            x = child.matrix[4]
            child.matrix.set(x0=x + offset + (x - left) * factor)

    partition: relation_many[UML.ActivityPartition] = association(
        "partition", UML.ActivityPartition, composite=True
    )

    def load(self, name, value):
        self._loading = True
        return super().load(name, value)

    def postload(self):
        super().postload()
        if self.subject and self.subject not in self.partition:
            self.partition = self.subject
        self._loading = False

    def update_shapes(self, event=None) -> None:
        """Set the min width of all the swimlanes."""
        self.shape = Box(
            *(CssNode("swimlane", partition, Box()) for partition in self.partition),
            orientation=Orientation.HORIZONTAL,
            draw=self.draw_swimlanes,
        )

    def partition_at_point(self, pos: Pos) -> UML.ActivityPartition | None:
        if self.point(*pos) > 0:
            return None

        if partitions := self.partition:
            partition_width = self.width / len(partitions)
        else:
            partition_width = self.width / 2

        right = self.handles()[0].pos.x
        x, _y = pos
        for p in partitions:
            right += partition_width
            if x < right:
                return p
        return None

    def draw_swimlanes(
        self, box: Box, context: DrawContext, bounding_box: Rectangle
    ) -> None:
        """Draw the vertical partitions as connected swimlanes.

        The partitions are open on the bottom. We divide the total size
        by the total number of partitions and space them evenly.
        """
        cr = context.cairo
        cr.set_line_width(context.style.get("line-width", 2.4))
        partitions = self.partition

        if partitions:
            partition_width = bounding_box.width / len(partitions)
        else:
            partition_width = bounding_box.width / 2

        padding_top, padding_right, padding_bottom, padding_left = context.style.get(
            "padding", DEFAULT_PADDING
        )
        layout = Layout(
            font=context.style,
            width=partition_width - padding_left - padding_right,
        )
        header_height = 0.0
        for num, partition in enumerate(partitions):
            cr.move_to(partition_width * num, 0)
            cr.line_to(partition_width * num, bounding_box.height)
            layout.set(
                text=(
                    f"{partition.name}: {partition.represents.name}"
                    if isinstance(partition.represents, UML.NamedElement)
                    else partition.name
                )
            )
            cr.move_to(partition_width * num + padding_left, padding_top)
            layout.show_layout(cr)
            _, h = layout.size()
            header_height = max(h, header_height)

        cr.move_to(bounding_box.width, 0)
        cr.line_to(bounding_box.width, bounding_box.height)

        header_height += padding_top + padding_bottom
        cr.move_to(0, 0)
        cr.line_to(0 + bounding_box.width, 0)
        cr.move_to(0, header_height)
        cr.line_to(0 + bounding_box.width, header_height)
        stroke(context, fill=True)



from gaphor.diagram.shapes import Box, CssNode, Text, draw_top_separator


def region_compartment(subject):
    if not subject:
        return

    for index, region in enumerate(subject.region):
        yield _create_region_compartment(region, index)


def _create_region_compartment(region, index):
    return CssNode(
        "region",
        region,
        Box(
            Text(
                text=lambda: region.name or "",
            ),
            draw=draw_top_separator,
        ),
    )



from gaphor.core.modeling.properties import attribute
from gaphor.diagram.presentation import AttachedPresentation, Named
from gaphor.diagram.shapes import (
    Box,
    CssNode,
    IconBox,
    Text,
    TextAlign,
    VerticalAlign,
    draw_border,
)
from gaphor.diagram.support import represents
from gaphor.SysML import sysml
from gaphor.UML.compartments import text_stereotypes


def text_position(position):
    return {
        "text-align": TextAlign.LEFT if position == "left" else TextAlign.RIGHT,
        "vertical-align": VerticalAlign.BOTTOM
        if position == "bottom"
        else VerticalAlign.TOP,
    }


@represents(sysml.ProxyPort)
class ProxyPortItem(Named, AttachedPresentation[sysml.ProxyPort]):
    def __init__(self, diagram, id=None):
        super().__init__(diagram, id, width=16, height=16)
        self.watch("subject[UML:NamedElement].name").watch(
            "subject[UML:TypedElement].type.name"
        ).watch("show_type")

    show_type: attribute[int] = attribute("show_type", int, default=False)

    def update_shapes(self, event=None):
        self.shape = IconBox(
            Box(draw=draw_border),
            text_stereotypes(self, lambda: [self.diagram.gettext("proxy")]),
            CssNode(
                "name",
                self.subject,
                Text(
                    text=self._format_name,
                ),
            ),
        )

    def _format_name(self):
        if not self.subject:
            return ""

        name = self.subject.name or ""
        if self.show_type and self.subject.type:
            return f"{name}: {self.subject.type.name or ''}"
        return name
